#!/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from xpra.build_info import packages

from packageurl.contrib import url2purl

from cyclonedx.builder.this import this_component as cdx_lib_component
from cyclonedx.factory.license import LicenseFactory
from cyclonedx.model.bom import Bom
from cyclonedx.model.component import Component, ComponentType
from cyclonedx.output.json import JsonV1Dot5


lc_factory = LicenseFactory()

bom = Bom()
bom.metadata.tools.components.add(cdx_lib_component())
bom.metadata.tools.components.add(Component(
    name='my-own-SBOM-generator',
    type=ComponentType.APPLICATION,
))

bom.metadata.component = root_component = Component(
    name='Xpra',
    type=ComponentType.APPLICATION,
    licenses=[lc_factory.make_from_string('GPL2+')],
    bom_ref='Xpra',
)

# cache components so we can more easily add dependencies later:
components = {}
# resolve full mingw64 prefixed package names to short version:
names = {}
dependencies = {}
for package_name, info in packages.items():
    license_names = []
    licenses = [
        lc_factory.make_from_string(license_name)
        for license_name in [info.get("Licenses", "")]
    ]
    component = Component(
        type=ComponentType.LIBRARY,
        name=package_name,
        version=info.get("Version", ""),
        licenses=licenses,
        purl=url2purl.get_purl(info.get("URL", "")),
    )
    bom.components.add(component)
    components[package_name] = component
    names[info.get("Name", package_name)] = package_name
    provides = info.get("Provides", ())
    for provided in provides:
        if provided not in names:
            names[provided] = package_name
    if not info.get("Required By"):
        bom.register_dependency(root_component, [component])
    deps = info.get("Depends On")
    if deps:
        dependencies[package_name] = deps


# now resolve the intra component dependencies:
for package_name, deps in dependencies.items():
    component = components[package_name]
    dep_components = []
    for dep in deps:
        dep_package_name = names.get(dep, dep)
        dep_component = components.get(dep_package_name)
        if dep_component:
            dep_components.append(dep_component)
        else:
            if package_name.find("gst") >= 0:
                continue
            if any(dep.endswith(f"-{suffix}") for suffix in (
                # only needed at build time:
                "cc", "python-build", "python-installer", "python-packaging",
                # gtk dependencies we don't need:
                "gtk-update-icon-cache", "shared-mime-info", "json-glib",
                "adwaita-icon-theme",
                # python dependencies we don't want:
                "tcl", "tk", "headers", "tzdata",
                # python works fine without `mpdec` so perhaps the current builds of Python
                # do not yet use the unbundled mpdec library and the MSYS2 package dependencies are therefore wrong?
                # in any case, it will become a hard dependency in python 3.13: /mingw64/bin/libmpdec-4.dll
                # at which point this workaround can be removed:
                "mpdecimal",
                # webp tools can use this, but we don't ship them:
                "giflib",
                # cairo, why?
                "lzo2",
                # could this be due to outdated dependencies?
                # paramiko and pyu2f work just fine without it:
                "python-six",
                # libcurl may use them, instead we only rely on the OS:
                "ca-certificates",
                # we skip the gnome bits of glib-networking:
                "gsettings-desktop-schemas",
                # libproxy wants this package, but we don't:
                "duktape",
                # ncurses works just fine without it:
                "libsystre",
                # sqlite tools use this, but we don't ship them:
                "readline",
            )
            ):
                continue
            sys.stderr.write(f"Warning: {dep!r} not found, dependency of {package_name}\n")
    if dep_components:
        bom.register_dependency(component, dep_components)

json = JsonV1Dot5(bom).output_as_string(indent=2)

if len(sys.argv) > 1:
    filename = sys.argv[1]
    with open(filename, "w") as f:
        f.write(json)
else:
    print(json)



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from ctypes import CDLL, c_int, c_char_p

from xpra.util.objects import typedict
from xpra.os_util import POSIX
from xpra.util.str_fn import strtobytes
from xpra.auth.sys_auth_base import SysAuthenticator, log

LIBWRAP = os.environ.get("XPRA_LIBWRAP", "libwrap.so.0")

PRG_NAME = b"xpra"
prg = c_char_p(PRG_NAME)
UNKNOWN = b""
unknown = c_char_p(UNKNOWN)


def check_host(peername: str, host: str) -> bool:
    libwrap = CDLL(LIBWRAP)
    assert libwrap
    hosts_ctl = libwrap.hosts_ctl
    hosts_ctl.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p]
    hosts_ctl.restype = c_int

    log("check_host(%s, %s)", peername, host)
    # name = c_char_p(username)
    c_host = c_char_p(strtobytes(host))
    c_peername = c_char_p(strtobytes(peername))
    # v = hosts_ctl(prg, c_host, unknown, unknown)
    v = hosts_ctl(prg, c_peername, c_host, unknown)
    log("hosts_ctl%s=%s", (PRG_NAME, peername, host, UNKNOWN), v)
    return bool(v)


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        log("hosts.Authenticator(%s)", kwargs)
        if not POSIX:
            log.warn("Warning: hosts authentication is not supported on %s", os.name)
            return
        connection = kwargs.get("connection", None)
        try:
            from xpra.net.bytestreams import SocketConnection  # pylint: disable=import-outside-toplevel
            if not connection and isinstance(connection, SocketConnection):
                raise ValueError(f"hosts: invalid connection {connection!r} (not a socket connection)")
            info = connection.get_info()
            log("hosts.Authenticator(..) connection info=%s", info)
            host = info.get("remote")[0]
            peername = info.get("endpoint")[0]
        except Exception as e:
            log.error("Error: cannot get host from connection")
            log.estr(e)
            raise
        self.peername = peername
        self.host = host
        super().__init__(**kwargs)

    def requires_challenge(self) -> bool:
        return False

    def authenticate(self, _caps: typedict) -> bool:
        if not self.host or not check_host(self.peername, self.host):
            errinfo = "'%s'" % self.peername
            if self.peername != self.host:
                errinfo += " ('%s')" % self.host
            log.warn("Warning: access denied for host %s" % errinfo)
            return False
        return True

    def __repr__(self):
        return "hosts"


def main(argv) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.log import consume_verbose_argv
    from xpra.platform import program_context
    with program_context("Host Check", "Host Check"):
        consume_verbose_argv(argv, "auth")
        if len(argv) < 3:
            print("usage: %s peername1 hostname1 [peername2 hostname2] [..]" % sys.argv[0])
            return 1
        argv = argv[1:]
        while len(argv) >= 2:
            peername, host = argv[:2]
            check = check_host(peername, host)
            print(f"host check for {peername!r}, {host!r}: {check}")
            argv = argv[2:]
    return 0


if __name__ == "__main__":
    r = main(sys.argv)
    sys.exit(r)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Sequence, Callable

from xpra.server import features
from xpra.util.str_fn import print_nested_dict
from xpra.util.objects import typedict, merge_dicts
from xpra.common import LOG_HELLO
from xpra.log import Logger

log = Logger("server")


def get_enabled_mixins() -> Sequence[type]:
    # pylint: disable=import-outside-toplevel
    from xpra.server.source.clientinfo import ClientInfoMixin
    mixins: list[type] = [ClientInfoMixin]
    if features.notifications:
        from xpra.server.source.notification import NotificationMixin
        mixins.append(NotificationMixin)
    if features.clipboard:
        from xpra.server.source.clipboard import ClipboardConnection
        mixins.append(ClipboardConnection)
    if features.audio:
        from xpra.server.source.audio import AudioMixin
        mixins.append(AudioMixin)
    if features.webcam:
        from xpra.server.source.webcam import WebcamMixin
        mixins.append(WebcamMixin)
    if features.fileprint:
        from xpra.server.source.fileprint import FilePrintMixin
        mixins.append(FilePrintMixin)
    if features.mmap:
        from xpra.server.source.mmap import MMAP_Connection
        mixins.append(MMAP_Connection)
    if features.input_devices:
        from xpra.server.source.input import InputMixin
        mixins.append(InputMixin)
    if features.dbus:
        from xpra.server.source.dbus import DBUS_Mixin
        mixins.append(DBUS_Mixin)
    if features.network_state:
        from xpra.server.source.networkstate import NetworkStateMixin
        mixins.append(NetworkStateMixin)
    if features.shell:
        from xpra.server.source.shell import ShellMixin
        mixins.append(ShellMixin)
    if features.display:
        from xpra.server.source.display import ClientDisplayMixin
        mixins.append(ClientDisplayMixin)
    if features.cursors:
        from xpra.server.source.cursors import CursorsMixin
        mixins.append(CursorsMixin)
    if features.windows:
        from xpra.server.source.windows import WindowsMixin
        mixins.append(WindowsMixin)
    # must be after windows mixin, so that it can assume "self.send_windows" is set
    if features.encoding:
        from xpra.server.source.encodings import EncodingsMixin
        mixins.append(EncodingsMixin)
    if features.audio and features.av_sync:
        from xpra.server.source.avsync import AVSyncMixin
        mixins.append(AVSyncMixin)
    from xpra.server.source.idle_mixin import IdleMixin
    mixins.append(IdleMixin)
    return tuple(mixins)


def get_needed_based_classes(caps: typedict) -> tuple[type, ...]:
    from xpra.server.source.client_connection import ClientConnection
    classes: list[type] = [ClientConnection]
    mixins = get_enabled_mixins()
    for c in mixins:
        r = c.is_needed(caps)
        log("get_client_connection_class(..) %s enabled=%s", c.__name__.split(".")[-1], r)
        if r:
            classes.append(c)
    return tuple(classes)


def get_client_connection_class(caps: typedict):
    CC_BASES = get_needed_based_classes(caps)
    ClientConnectionClass = type('ClientConnectionClass', CC_BASES, {})
    log("ClientConnectionClass%s", CC_BASES)

    class ClientConnectionMuxer(ClientConnectionClass):

        def __init__(self, protocol, disconnect_cb: Callable, server, setting_changed: Callable):
            self.hello_sent = False
            from xpra.server.source.client_connection import ClientConnection
            for bc in CC_BASES:
                try:
                    if bc == ClientConnection:
                        initargs = (protocol, disconnect_cb, setting_changed)
                    else:
                        initargs = ()
                    bc.__init__(self, *initargs)
                    bc.init_from(self, protocol, server)
                except Exception as e:
                    log.error("%s.__init__(..)", bc, exc_info=True)
                    raise RuntimeError(f"failed to initialize {bc}: {e}") from None

            for c in CC_BASES:
                c.init_state(self)
            self.run()

        def close(self) -> None:
            log("%s.close()", self)
            for bc in reversed(CC_BASES):
                log("%s.cleanup()", bc)
                try:
                    bc.cleanup(self)
                except Exception as e:
                    log("%s.cleanup()", bc, exc_info=True)
                    log.error("Error closing connection,")
                    log.error(" %s in %s module:", type(e).__name__, bc.__name__)
                    log.estr(e)
                    raise RuntimeError(f"failed to close {bc}: {e}") from None

        def send_hello(self, server_capabilities: dict) -> None:
            capabilities = server_capabilities.copy()
            for bc in CC_BASES:
                caps = bc.get_caps(self)
                log("%s.get_caps()=%s", bc, caps)
                merge_dicts(capabilities, caps)
            if LOG_HELLO:
                netlog = Logger("network")
                netlog.info(f"sending hello to {self}:")
                print_nested_dict(capabilities, print_fn=netlog.info)
            self.send("hello", capabilities)
            self.hello_sent = True

        def threaded_init_complete(self, server) -> None:
            log("threaded_init_complete(%s) calling base classes %s", server, CC_BASES)
            for bc in CC_BASES:
                log("calling %s.threaded_init_complete(%s, %s)", bc, self, server)
                bc.threaded_init_complete(self, server)

        def suspend(self) -> None:
            log("suspend()")
            for bc in CC_BASES:
                bc.suspend(self)

        def resume(self) -> None:
            log("resume()")
            for bc in CC_BASES:
                bc.resume(self)

        def get_info(self) -> dict[str, Any]:
            def module_name(m) -> str:
                name = str(m.__name__.split(".")[-1])
                return name.replace("Mixin", "").replace("Connection", "").rstrip("_")

            info = {
                "modules": tuple(module_name(x) for x in CC_BASES),
            }
            for bc in CC_BASES:
                log("%s.get_info()", bc)
                try:
                    merge_dicts(info, bc.get_info(self))
                except Exception as e:
                    log("merge_dicts on %s", bc, exc_info=True)
                    log.error("Error: cannot add information from %s:", bc)
                    log.estr(e)
            return info

        def parse_hello(self, c: typedict) -> None:
            self.ui_client = c.boolget("ui_client", True)
            self.wants: list[str] = list(c.strtupleget("wants", self.wants))
            for x, enabled in {
                "encodings": self.ui_client,
                "display": self.ui_client,
                "versions": True,
                "features": True,
            }.items():
                if enabled:
                    self.wants.append(x)
            for bc in CC_BASES:
                log("%s.parse_client_caps(..)", bc)
                bc.parse_client_caps(self, c)
            # log client info:
            cinfo = self.get_connect_info()
            for i, ci in enumerate(cinfo):
                log.info("%s%s", ["", " "][int(i > 0)], ci)
            if self.client_proxy:
                from xpra.util.version import version_compat_check
                msg = version_compat_check(self.proxy_version)
                if msg:
                    proxylog = Logger("proxy")
                    proxylog.warn("Warning: proxy version may not be compatible: %s", msg)

    return ClientConnectionMuxer



# This file is part of Xpra.
# Copyright (C) 2013 Arthur Huillet
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False

from time import monotonic
from typing import Any, Tuple, Dict
from collections.abc import Sequence

from xpra.log import Logger
log = Logger("csc", "libyuv")

from xpra.util.str_fn import csv
from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.codecs.constants import get_subsampling_divs, CSCSpec
from xpra.codecs.image import ImageWrapper
from xpra.buffers.membuf cimport getbuf, MemBuf, memalign, buffer_context    # pylint: disable=syntax-error

from xpra.codecs.argb.argb cimport show_plane_range
from libc.stdint cimport uint8_t, uintptr_t
from libc.stdlib cimport free


cdef uint8_t SHOW_PLANE_RANGES = envbool("XPRA_SHOW_PLANE_RANGES", False)


cdef extern from "Python.h":
    object PyMemoryView_FromMemory(char *mem, Py_ssize_t size, int flags)
    int PyBUF_WRITE

cdef extern from "libyuv/convert_argb.h" namespace "libyuv":
    int YUY2ToARGB(const uint8_t* src_yuy2,
               int src_stride_yuy2,
               uint8_t* dst_argb,
               int dst_stride_argb,
               int width,
               int height) nogil

cdef extern from "libyuv/convert_from_argb.h" namespace "libyuv":
    #int BGRAToI420(const uint8_t* src_frame, ...
    #this is actually BGRX for little endian systems:
    int ARGBToI420(const uint8_t* src_frame, int src_stride_frame,
               uint8_t* dst_y, int dst_stride_y,
               uint8_t* dst_u, int dst_stride_u,
               uint8_t* dst_v, int dst_stride_v,
               int width, int height) nogil

    int ARGBToJ420(const uint8_t* src_frame, int src_stride_frame,
               uint8_t* dst_y, int dst_stride_y,
               uint8_t* dst_u, int dst_stride_u,
               uint8_t* dst_v, int dst_stride_v,
               int width, int height) nogil

    int ARGBToNV12(const uint8_t* src_argb, int src_stride_argb,
                   uint8_t* dst_y, int dst_stride_y,
                   uint8_t* dst_uv, int dst_stride_uv,
                   int width, int height) nogil

    int NV12ToRGB24(const uint8_t* src_y, int src_stride_y,
                    const uint8_t* src_uv, int src_stride_uv,
                    uint8_t* dst_rgb24, int dst_stride_rgb24,
                    int width, int height) nogil

    int NV12ToARGB(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_uv, int src_stride_uv,
                   uint8_t* dst_argb, int dst_stride_argb,
                   int width, int height) nogil

    int NV12ToABGR(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_uv,int src_stride_uv,
                   uint8_t* dst_abgr, int dst_stride_abgr,
                   int width, int height) nogil

    int I420ToRGB24(const uint8_t* src_y, int src_stride_y,
                    const uint8_t* src_u, int src_stride_u,
                    const uint8_t* src_v, int src_stride_v,
                    uint8_t* dst_rgb24, int dst_stride_rgb24,
                    int width, int height) nogil

    int J420ToRGB24(const uint8_t* src_y, int src_stride_y,
                    const uint8_t* src_u, int src_stride_u,
                    const uint8_t* src_v, int src_stride_v,
                    uint8_t* dst_rgb24, int dst_stride_rgb24,
                    int width, int height) nogil

    int I420ToRGBA(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_u, int src_stride_u,
                   const uint8_t* src_v, int src_stride_v,
                   uint8_t* dst_rgba, int dst_stride_rgba,
                   int width, int height) nogil

    int I420ToABGR(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_u, int src_stride_u,
                   const uint8_t* src_v, int src_stride_v,
                   uint8_t* dst_abgr, int dst_stride_abgr,
                   int width, int height) nogil

    int J420ToABGR(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_u, int src_stride_u,
                   const uint8_t* src_v, int src_stride_v,
                   uint8_t* dst_abgr, int dst_stride_abgr,
                   int width, int height) nogil

    int I420ToARGB(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_u, int src_stride_u,
                   const uint8_t* src_v, int src_stride_v,
                   uint8_t* dst_abgr, int dst_stride_abgr,
                   int width, int height) nogil

    int J420ToARGB(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_u, int src_stride_u,
                   const uint8_t* src_v, int src_stride_v,
                   uint8_t* dst_abgr, int dst_stride_abgr,
                   int width, int height) nogil

    int I444ToARGB(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_u, int src_stride_u,
                   const uint8_t* src_v, int src_stride_v,
                   uint8_t* dst_argb, int dst_stride_argb,
                   int width, int height) nogil

    int I444ToABGR(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_u, int src_stride_u,
                   const uint8_t* src_v, int src_stride_v,
                   uint8_t* dst_abgr, int dst_stride_abgr,
                   int width, int height) nogil

    int J444ToARGB(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_u, int src_stride_u,
                   const uint8_t* src_v, int src_stride_v,
                   uint8_t* dst_argb, int dst_stride_argb,
                   int width, int height) nogil

    int J444ToABGR(const uint8_t* src_y, int src_stride_y,
                   const uint8_t* src_u, int src_stride_u,
                   const uint8_t* src_v, int src_stride_v,
                   uint8_t* dst_abgr, int dst_stride_abgr,
                   int width, int height) nogil

    int ARGBToI444(const uint8_t* src_argb, int src_stride_argb,
                   uint8_t* dst_y, int dst_stride_y,
                   uint8_t* dst_u, int dst_stride_u,
                   uint8_t* dst_v, int dst_stride_v,
                   int width, int height) nogil


cdef extern from "libyuv/scale.h" namespace "libyuv":
    ctypedef enum FilterMode:
        kFilterNone
        kFilterBilinear
        kFilterBox
    void ScalePlane(const uint8_t* src, int src_stride,
                int src_width, int src_height,
                uint8_t* dst, int dst_stride,
                int dst_width, int dst_height,
                FilterMode filtering) nogil

cdef extern from "libyuv/scale_argb.h" namespace "libyuv":
    int ARGBScale(const uint8_t* src_argb,
              int src_stride_argb,
              int src_width,
              int src_height,
              uint8_t* dst_argb,
              int dst_stride_argb,
              int dst_width,
              int dst_height,
              FilterMode filtering) nogil

cdef extern from "libyuv/planar_functions.h" namespace "libyuv":
    int ARGBGrayTo(const uint8_t* src_argb,
                   int src_stride_argb,
                   uint8_t* dst_argb,
                   int dst_stride_argb,
                   int width,
                   int height) nogil

cdef inline str get_fiter_mode_str(FilterMode fm):
    if fm==kFilterNone:
        return "None"
    elif fm==kFilterBilinear:
        return "Bilinear"
    elif fm==kFilterBox:
        return  "Box"
    return "invalid"


cdef inline FilterMode get_filtermode(int speed, int downscaling):
    if downscaling:
        return kFilterBilinear
    if speed>66:
        return kFilterNone
    if speed>33:
        return kFilterBilinear
    return kFilterBox


DEF ALIGN = 4   #MEMALIGN_ALIGNMENT


cdef inline int roundup(int n, int m):
    return (n + m - 1) & ~(m - 1)


cdef inline uintptr_t roundupl(uintptr_t n, uintptr_t m):
    return (n + m - 1) & ~(m - 1)


cdef inline uintptr_t memalign_ptr(uintptr_t ptr):
    return <uintptr_t> roundupl(<uintptr_t> ptr, ALIGN)


def get_type() -> str:
    return "libyuv"


def get_version() -> Tuple[int, int]:
    return (1, 0)


#hardcoded for now:
MAX_WIDTH = 32768
MAX_HEIGHT = 32768
COLORSPACES: Dict[str, Sequence[str]] = {
    "BGRX" : ("YUV444P", "YUV420P", "NV12"),
    "NV12" : ("RGB", "BGRX", "RGBX"),
    "YUV420P" : ("RGB", "XBGR", "RGBX", "BGRX"),
    "YUV444P" : ("BGRX", "RGBX"),
    "YUYV" : ("BGRX", ),
}


def get_info() -> Dict[str, Any]:
    return {
        "version"           : get_version(),
        "formats"           : COLORSPACES,
        "max-size"          : (MAX_WIDTH, MAX_HEIGHT),
    }


def get_specs() -> Sequence[CSCSpec]:
    specs: Sequence[CSCSpec] = []
    for in_cs, out_css in COLORSPACES.items():
        can_scale = in_cs not in ("NV12", "YUYV")
        specs.append(CSCSpec(
                input_colorspace=in_cs, output_colorspaces=out_css,
                codec_class=Converter, codec_type=get_type(),
                quality=100, speed=100,
                setup_cost=0, min_w=8, min_h=2, can_scale=can_scale,
                max_w=MAX_WIDTH, max_h=MAX_HEIGHT,
                )
            )
    return specs


class YUVImageWrapper(ImageWrapper):

    def _cn(self) -> str:
        return "libyuv.YUVImageWrapper"

    def free(self) -> None:
        cdef uintptr_t buf = self.cython_buffer
        self.cython_buffer = 0
        log("libyuv.YUVImageWrapper.free() cython_buffer=%#x", buf)
        super().free()
        if buf!=0:
            free(<void *> buf)


def argb_to_gray(image: ImageWrapper) -> ImageWrapper:
    cdef iplanes = image.get_planes()
    pixels = image.get_pixels()
    cdef int stride = image.get_rowstride()
    cdef int width = image.get_width()
    cdef int height = image.get_height()
    assert iplanes==ImageWrapper.PACKED, "invalid plane input format: %s" % iplanes
    assert pixels, "failed to get pixels from %s" % image
    #allocate output buffer:
    cdef int dst_stride = width*4
    cdef MemBuf output_buffer = getbuf(dst_stride*height, 0)
    if not output_buffer:
        raise RuntimeError("failed to allocate %i bytes for output buffer" % (dst_stride*height))
    cdef uint8_t* buf = <uint8_t*> output_buffer.get_mem()
    cdef int result = -1
    cdef const uint8_t* src
    with buffer_context(pixels) as bc:
        src = <const uint8_t *> (<uintptr_t> int(bc))
        with nogil:
            result = ARGBGrayTo(src, stride,
                                buf, dst_stride,
                                width, height)
    assert result==0, "libyuv ARGBGrayTo failed and returned %i" % result
    out = memoryview(output_buffer)
    gray_image = ImageWrapper(0, 0, width, height, out, image.get_pixel_format(), 24, dst_stride, image.get_bytesperpixel(), ImageWrapper.PACKED)
    log("argb_to_gray(%s)=%s", image, gray_image)
    return gray_image


def argb_scale(image: ImageWrapper, int src_width, int src_height,
               int dst_width, int dst_height, FilterMode filtermode=kFilterNone) -> ImageWrapper:
    cdef iplanes = image.get_planes()
    pixels = image.get_pixels()
    cdef int stride = image.get_rowstride()
    cdef int width = image.get_width()
    cdef int height = image.get_height()
    if width<src_width or height<src_height:
        raise ValueError("image is too small to be scaled: %ix%i, expected at least %ix%i" % (
            width, height, src_width, src_height,
        ))
    cdef int bpp = image.get_bytesperpixel()
    assert bpp in (3, 4), "invalid bytes per pixel: %s" % bpp
    assert iplanes==ImageWrapper.PACKED, "invalid plane input format: %s" % iplanes
    assert pixels, "failed to get pixels from %s" % image
    #allocate output buffer:
    cdef int dst_stride = dst_width*4
    cdef MemBuf output_buffer = getbuf(dst_stride*dst_height, 0)
    if not output_buffer:
        raise RuntimeError("failed to allocate %i bytes for output buffer" % (dst_stride*height))
    cdef uint8_t* buf = <uint8_t*> output_buffer.get_mem()
    cdef int result = -1
    cdef const uint8_t* src
    with buffer_context(pixels) as bc:
        src = <const uint8_t *> (<uintptr_t> int(bc))
        with nogil:
            result = ARGBScale(src,
                               stride, src_width, src_height,
                               buf, dst_stride, dst_width, dst_height,
                               filtermode)
    assert result==0, "libyuv ARGBScale failed and returned %i" % result
    out = memoryview(output_buffer)
    scaled_image = ImageWrapper(0, 0, dst_width, dst_height, out,
                                image.get_pixel_format(), image.get_depth(), dst_stride, bpp, ImageWrapper.PACKED)
    log("argb_scale(%s, %i, %i, %i)=%s", image, dst_width, dst_height, filtermode, scaled_image)
    return scaled_image


cdef class Converter:
    cdef int src_width
    cdef int src_height
    cdef int dst_width
    cdef int dst_height
    cdef int dst_full_range
    cdef uint8_t yuv_scaling
    cdef uint8_t rgb_scaling
    cdef int planes

    cdef unsigned long frames
    cdef double time

    cdef object src_format
    cdef object dst_format
    cdef int out_stride[3]
    cdef int out_width[3]
    cdef int out_height[3]
    cdef unsigned long[3] out_offsets
    cdef unsigned long out_size[3]
    cdef unsigned long out_buffer_size
    #when yuv_scaling:
    cdef uint8_t *output_buffer
    cdef FilterMode filtermode
    cdef int scaled_stride[3]
    cdef int scaled_width[3]
    cdef int scaled_height[3]
    cdef unsigned long[3] scaled_offsets
    cdef unsigned long scaled_size[3]
    cdef unsigned long scaled_buffer_size

    cdef object __weakref__

    def init_context(self, int src_width, int src_height, src_format: str,
                           int dst_width, int dst_height, dst_format: str,
                           options: typedict) -> None:
        log("libyuv.Converter.init_context%s", (
            src_width, src_height, src_format, dst_width, dst_height, dst_format, options))
        if src_format not in COLORSPACES:
            raise ValueError(f"invalid input colorspace: {src_format}, must be one of " + csv(COLORSPACES.keys()))
        if dst_format not in COLORSPACES[src_format]:
            raise ValueError(f"invalid output colorspace {dst_format} for {src_format} input, must be one of " + csv(COLORSPACES.get(src_format, ())))
        self.src_format = src_format
        self.dst_format = dst_format
        cdef int speed = options.intget("speed", 100)
        cdef int downscaling = dst_width < src_width or dst_height < src_height
        self.filtermode = get_filtermode(speed, downscaling)
        self.src_width = src_width
        self.src_height = src_height
        self.dst_width = dst_width
        self.dst_height = dst_height
        self.dst_full_range = options.boolget("full-range")
        self.out_buffer_size = 0
        self.scaled_buffer_size = 0
        self.time = 0
        self.frames = 0
        self.output_buffer = NULL
        self.yuv_scaling = False
        self.rgb_scaling = False
        cdef uint8_t scaling = int(src_width!=dst_width or src_height!=dst_height)
        if dst_format=="YUV420P" or dst_format=="YUV444P":
            self.planes = 3
            self.yuv_scaling = scaling
            if scaling:
                self.init_yuv_scaled_buffer(dst_format, dst_width, dst_height)
            self.init_yuv_output()
        elif dst_format=="NV12":
            self.planes = 2
            self.rgb_scaling = scaling
            self.init_yuv_output()
        elif dst_format in ("RGB", "BGRX", "RGBX", "XBGR"):
            self.planes = 1
            self.yuv_scaling = scaling
            if scaling:
                self.init_yuv_scaled_buffer(src_format, dst_width, dst_height)
            self.out_buffer_size = dst_width*len(dst_format)*dst_height
        elif src_format == "YUYV":
            assert dst_format in ("BGRX", )
            assert not scaling
            self.planes = 1
            self.out_buffer_size = dst_width * 4 * dst_height
        else:
            raise ValueError(f"invalid destination format: {dst_format!r}")
        log(f"{src_format} -> {dst_format} planes={self.planes}, output buffer-size={self.out_buffer_size}")
        log(f" yuv-scaling={self.yuv_scaling}, rgb-scaling={self.rgb_scaling}, full-range={self.dst_full_range}")

    def init_yuv_output(self) -> None:
        #pre-calculate unscaled YUV plane heights:
        divs = get_subsampling_divs(self.dst_format)
        for i in range(self.planes):
            xdiv, ydiv = divs[i]
            if self.rgb_scaling:
                #we scale before csc to the dst size:
                self.out_width[i]   = self.dst_width // xdiv
                self.out_height[i]  = self.dst_height // ydiv
            else:
                #we don't scale, so the size is the src size:
                self.out_width[i]   = self.src_width // xdiv
                self.out_height[i]  = self.src_height // ydiv
            self.out_stride[i]  = roundup(self.out_width[i], ALIGN)
            self.out_size[i]    = self.out_stride[i] * self.out_height[i]
            self.out_offsets[i] = self.out_buffer_size
            #add two extra lines to height so we can access two rowstrides at a time,
            #no matter where we start to read on the last line
            #and round up to memalign each plane:
            #(why two and not just one? libyuv will do this for input data with odd height)
            self.out_buffer_size += roundupl(self.out_size[i] + 2*self.out_stride[i], ALIGN)
        if self.yuv_scaling:
            #re-use the same temporary buffer every time before scaling:
            self.output_buffer = <uint8_t *> memalign(self.out_buffer_size)
            if self.output_buffer==NULL:
                raise RuntimeError("failed to allocate %i bytes for output buffer" % self.out_buffer_size)
        log("buffer size=%i, yuv_scaling=%s, rgb_scaling=%s, filtermode=%s",
            self.out_buffer_size, self.yuv_scaling, self.rgb_scaling, get_fiter_mode_str(self.filtermode))

    def init_yuv_scaled_buffer(self, yuv_format, int width, int height) -> None:
        assert self.yuv_scaling
        self.scaled_buffer_size = 0
        divs = get_subsampling_divs(yuv_format)
        cdef int nplanes = len(divs)
        assert 0 < nplanes <= 3
        for i in range(nplanes):
            xdiv, ydiv = divs[i]
            self.scaled_width[i]    = width // xdiv
            self.scaled_height[i]   = height // ydiv
            self.scaled_stride[i]   = roundup(self.scaled_width[i], ALIGN)
            self.scaled_size[i]     = self.scaled_stride[i] * self.scaled_height[i]
            self.scaled_offsets[i]  = self.scaled_buffer_size
            self.scaled_buffer_size += self.scaled_size[i] + self.out_stride[i]

    def get_info(self) -> Dict[str,Any]:
        info = get_info()
        info |= {
            "frames"    : int(self.frames),
            "src_width" : self.src_width,
            "src_height": self.src_height,
            "dst_width" : self.dst_width,
            "dst_height": self.dst_height,
            "planes"    : self.planes,
        }
        if self.yuv_scaling:
            info["yuv-scaling"] = True
        if self.rgb_scaling:
            info["rgb-scaling"] = True
        if self.src_format:
            info["src_format"] = self.src_format
        if self.dst_format:
            info["dst_format"] = self.dst_format
        if self.frames>0 and self.time>0:
            pps = self.src_width * self.src_height * self.frames / self.time
            info["total_time_ms"] = int(self.time * 1000)
            info["pixels_per_second"] = int(pps)
        return info

    def __repr__(self):
        if not self.src_format or not self.dst_format:
            return "libyuv(uninitialized)"
        return "libyuv(%s %sx%s %s)" % (self.src_format, self.src_width, self.src_height, self.dst_format)

    def __dealloc__(self):
        self.clean()

    def get_src_width(self) -> int:
        return self.src_width

    def get_src_height(self) -> int:
        return self.src_height

    def get_src_format(self) -> str:
        return self.src_format

    def get_dst_width(self) -> int:
        return self.dst_width

    def get_dst_height(self) -> int:
        return self.dst_height

    def get_dst_format(self) -> str:
        return self.dst_format

    def get_type(self) -> str:
        return  "libyuv"

    def clean(self) -> None:
        self.src_width = 0
        self.src_height = 0
        self.dst_width = 0
        self.dst_height = 0
        self.src_format = ""
        self.dst_format = ""
        self.frames = 0
        self.time = 0
        self.yuv_scaling = 0
        self.rgb_scaling = 0
        for i in range(3):
            self.out_stride[i] = 0
            self.out_size[i] = 0
        cdef uint8_t *output_buffer = self.output_buffer
        self.output_buffer = NULL
        if output_buffer:
            free(output_buffer)
        self.out_buffer_size = 0

    def is_closed(self) -> bool:
        return self.out_buffer_size==0

    def convert_image(self, image: ImageWrapper) -> ImageWrapper:
        cdef int width = image.get_width()
        cdef int height = image.get_height()
        if width<self.src_width:
            raise ValueError(f"invalid image width: {width} (minimum is {self.src_width})")
        if height<self.src_height:
            raise ValueError(f"invalid image height: {height} (minimum is {self.src_height})")

        if self.src_format in ("BGRX", "BGRA"):
            return self.convert_bgrx_image(image)
        if self.src_format=="NV12":
            return self.convert_nv12_image(image)
        if self.src_format in ("YUV420P", "YUV444P"):
            return self.convert_yuv_image(image)
        if self.src_format == "YUYV":
            return self.convert_yuyv_image(image)
        raise RuntimeError(f"invalid source format {self.src_format}")

    def convert_nv12_image(self, image: ImageWrapper) -> ImageWrapper:
        self.validate_image_dst_size(image)
        assert not self.yuv_scaling
        cdef double start = monotonic()
        cdef int iplanes = image.get_planes()
        cdef int width = self.dst_width
        cdef int height = self.dst_height
        if iplanes!=2:
            raise ValueError(f"invalid number of planes: {iplanes} for {self.src_format}")
        if self.dst_format not in ("RGB", "BGRX", "RGBX"):
            raise ValueError(f"invalid dst format {self.dst_format!r}")
        if self.rgb_scaling:
            raise ValueError(f"cannot scale {self.src_format!r}")
        pixels = image.get_pixels()
        strides = image.get_rowstride()
        cdef int y_stride = strides[0]
        cdef int uv_stride = strides[1]
        cdef int Bpp = len(self.dst_format)
        cdef int rowstride = self.dst_width*Bpp
        cdef MemBuf rgb_buffer = getbuf(rowstride*height, 0)
        cdef uintptr_t y, uv
        cdef uint8_t *rgb
        cdef int r = 0
        rgb_buffer = getbuf(self.out_buffer_size, 0)
        if not rgb_buffer:
            raise RuntimeError(f"failed to allocate {self.out_buffer_size} bytes for output buffer")
        log("convert_nv12_image(%s) to %s", image, self.dst_format)
        if SHOW_PLANE_RANGES:
            show_plane_range("Y", pixels[0], width, y_stride, height)
            show_plane_range("UV", pixels[1], width, uv_stride, height//2)

        fn_name = "error"
        with buffer_context(pixels[0]) as y_buf:
            y = <uintptr_t> int(y_buf)
            with buffer_context(pixels[1]) as uv_buf:
                uv = <uintptr_t> int(uv_buf)
                rgb = <uint8_t*> rgb_buffer.get_mem()
                if self.dst_format=="RGB":
                    fn_name = "NV12ToRGB24"
                    with nogil:
                        r = NV12ToRGB24(<const uint8_t*> y, y_stride,
                                        <const uint8_t*> uv, uv_stride,
                                        rgb, rowstride,
                                        width, height)
                elif self.dst_format=="BGRX":
                    fn_name = "NV12ToARGB"
                    with nogil:
                        r = NV12ToARGB(<const uint8_t*> y, y_stride,
                                        <const uint8_t*> uv, uv_stride,
                                        rgb, rowstride,
                                        width, height)
                elif self.dst_format=="RGBX":
                    fn_name = "NV12ToABGR"
                    with nogil:
                        r = NV12ToABGR(<const uint8_t*> y, y_stride,
                                        <const uint8_t*> uv, uv_stride,
                                        rgb, rowstride,
                                        width, height)
                else:
                    raise RuntimeError(f"unexpected dst format {self.dst_format}")
        if r!=0:
            raise RuntimeError(f"libyuv NV12ToRGB failed and returned {r}")
        cdef double elapsed = monotonic()-start
        log("libyuv.%s took %.1fms", fn_name, 1000.0*elapsed)
        self.time += elapsed
        return ImageWrapper(0, 0, self.dst_width, self.dst_height,
                            rgb_buffer, self.dst_format, Bpp*8, rowstride, Bpp, ImageWrapper.PACKED)

    def validate_image_dst_size(self, image: ImageWrapper) -> None:
        # ensure that the image fits:
        # (it can be bigger if padded)
        if image.get_width() < self.dst_width or image.get_height() < self.dst_height:
            raise ValueError("image is too small for target: expected at least %ix%i but got %ix%i" % (
                self.dst_width, self.dst_height, image.get_width(), image.get_height(),
            ))

    def validate_image_src_size(self, image: ImageWrapper) -> None:
        # ensure that the image fits:
        # (it can be bigger if padded)
        if image.get_width() < self.src_width or image.get_height() < self.src_height:
            raise ValueError("image source is too small for source: expected at least %ix%i but got %ix%i" % (
                self.src_width, self.src_height, image.get_width(), image.get_height(),
            ))

    def scale_yuv_image(self, image: ImageWrapper) -> ImageWrapper:
        self.validate_image_src_size(image)
        assert self.scaled_buffer_size
        start = monotonic()
        cdef uint8_t *scaled_buffer = <unsigned char*> memalign(self.scaled_buffer_size)
        if scaled_buffer==NULL:
            raise RuntimeError(f"failed to allocate {self.scaled_buffer_size} bytes for scaled buffer")
        cdef uint8_t *scaled_planes[3]
        for i in range(self.planes):
            scaled_planes[i] = scaled_buffer + self.scaled_offsets[i]

        pixels = image.get_pixels()
        strides = image.get_rowstride()
        yuv_format = image.get_pixel_format()
        cdef int nplanes = image.get_planes()
        divs = get_subsampling_divs(yuv_format)

        cdef int src_width = 0
        cdef int src_height = 0
        cdef uintptr_t plane
        cdef int stride = 0

        # scale each plane:
        for i in range(nplanes):
            stride = strides[i]
            x_div, y_div = divs[i]
            src_width = self.src_width // x_div
            src_height = self.src_height // y_div
            with buffer_context(pixels[i]) as plane_buffer:
                plane = <uintptr_t> int(plane_buffer)
                scaled_planes[i] = scaled_buffer + self.scaled_offsets[i]
                #log(f"scale_yuv_image({image}) {stride=}, {x_div=}, {y_div=}, {src_width=}, {src_height=}")
                with nogil:
                    ScalePlane(<uint8_t*> plane, stride,
                               src_width, src_height,
                               scaled_planes[i], self.scaled_stride[i],
                               self.scaled_width[i], self.scaled_height[i],
                               self.filtermode)

        elapsed = monotonic()-start
        log("libyuv.ScalePlane %i planes, took %.1fms", nplanes, 1000.0*elapsed)
        strides = []
        planes = []
        for i in range(nplanes):
            strides.append(self.scaled_stride[i])
            planes.append(PyMemoryView_FromMemory(<char *> scaled_planes[i], self.scaled_size[i], PyBUF_WRITE))
        out_image = YUVImageWrapper(0, 0, self.dst_width, self.dst_height, planes, self.dst_format, 24, strides, 1, nplanes)
        out_image.set_full_range(image.get_full_range())
        out_image.cython_buffer = <uintptr_t> scaled_buffer
        return out_image

    def convert_yuyv_image(self, image: ImageWrapper) -> ImageWrapper:
        self.validate_image_dst_size(image)
        assert not self.yuv_scaling
        assert image.get_pixel_format() == "YUYV"
        cdef double start = monotonic()
        cdef iplanes = image.get_planes()
        assert iplanes == 0, f"expected packed pixels, got {iplanes}"
        if self.dst_format != "BGRX":
            raise ValueError(f"invalid dst format {self.dst_format!r}")

        cdef int width = self.dst_width
        cdef int height = self.dst_height
        if width < self.src_width:
            raise ValueError(f"source image width {width} is smaller than context {self.src_width}")
        if height < self.src_height:
            raise ValueError(f"source image height {height} is smaller than context {self.src_height}")
        assert self.dst_width >= self.src_width and self.dst_height >= self.src_height
        pixels = image.get_pixels()
        cdef int stride = image.get_rowstride()
        cdef int dst_stride = self.dst_width * 4

        rgb_buffer = getbuf(self.out_buffer_size, 0)
        if not rgb_buffer:
            raise RuntimeError(f"failed to allocate {self.out_buffer_size} bytes for output buffer")

        cdef uintptr_t yuyv
        cdef uint8_t *rgb
        cdef int r = 0
        with buffer_context(pixels) as yuyv_buf:
            yuyv = <uintptr_t> int(yuyv_buf)
            rgb = <uint8_t*> rgb_buffer.get_mem()
            with nogil:
                r = YUY2ToARGB(<const uint8_t*> yuyv,
                               stride,
                               rgb,
                               dst_stride,
                               width, height)
        if r!=0:
            raise RuntimeError(f"libyuv YUY2ToARGB failed and returned {r}")
        image.free()
        cdef double elapsed = monotonic()-start
        log("libyuv.YUY2ToARGB took %.1fms", 1000.0*elapsed)
        self.time += elapsed
        img = ImageWrapper(0, 0, self.dst_width, self.dst_height,
                           rgb_buffer, self.dst_format, 24, dst_stride, 4, ImageWrapper.PACKED)
        return img

    def convert_yuv_image(self, image: ImageWrapper) -> ImageWrapper:
        if self.yuv_scaling:
            image = self.scale_yuv_image(image)
        self.validate_image_dst_size(image)
        cdef double start = monotonic()
        cdef int iplanes = image.get_planes()
        cdef int width = self.dst_width
        cdef int height = self.dst_height
        cdef int full_range = image.get_full_range()
        if iplanes!=3:
            raise ValueError(f"invalid number of planes: {iplanes} for {self.src_format}")
        if self.src_format not in ("YUV420P", "YUV444P"):
            raise ValueError(f"invalid src format {self.src_format!r}")
        if self.dst_format not in ("RGB", "XBGR", "RGBX", "BGRX"):
            raise ValueError(f"invalid dst format {self.dst_format!r}")
        if self.rgb_scaling:
            raise ValueError(f"cannot scale {self.src_format}")
        pixels = image.get_pixels()
        strides = image.get_rowstride()
        cdef int y_stride = strides[0]
        cdef int u_stride = strides[1]
        cdef int v_stride = strides[2]
        cdef int Bpp = len(self.dst_format)
        cdef int rowstride = self.dst_width*Bpp
        cdef MemBuf rgb_buffer = getbuf(rowstride*height, 0)
        cdef uintptr_t y, u, v
        cdef uint8_t *rgb
        cdef int r = 0
        cdef int matched = 0
        rgb_buffer = getbuf(self.out_buffer_size, 0)
        if not rgb_buffer:
            raise RuntimeError(f"failed to allocate {self.out_buffer_size} bytes for output buffer")
        log("convert_yuv_image(%s) to %s", image, self.dst_format)
        if SHOW_PLANE_RANGES:
            divs = get_subsampling_divs(self.src_format)
            for i in range(3):
                xdiv, ydiv = divs[i]
                show_plane_range("YUV"[i], pixels[i], width // xdiv, strides[i], height // ydiv)

        fn_name = ""
        with buffer_context(pixels[0]) as y_buf:
            y = <uintptr_t> int(y_buf)
            with buffer_context(pixels[1]) as u_buf:
                u = <uintptr_t> int(u_buf)
                with buffer_context(pixels[2]) as v_buf:
                    v = <uintptr_t> int(v_buf)
                    rgb = <uint8_t*> rgb_buffer.get_mem()
                    if self.dst_format=="RGB" and self.src_format=="YUV420P":
                        fn_name = "J420ToRGB24" if full_range else "I420ToRGB24"
                        with nogil:
                            if full_range:
                                r = J420ToRGB24(<const uint8_t*> y, y_stride,
                                                <const uint8_t*> u, u_stride,
                                                <const uint8_t*> v, v_stride,
                                                rgb, rowstride,
                                                width, height)
                            else:
                                r = I420ToRGB24(<const uint8_t*> y, y_stride,
                                                <const uint8_t*> u, u_stride,
                                                <const uint8_t*> v, v_stride,
                                                rgb, rowstride,
                                                width, height)
                    elif self.dst_format=="XBGR" and self.src_format=="YUV420P":
                        fn_name = "I420ToRGBA"
                        # we can't handle full-range here!
                        with nogil:
                            r = I420ToRGBA(<const uint8_t*> y, y_stride,
                                           <const uint8_t*> u, u_stride,
                                           <const uint8_t*> v, v_stride,
                                           rgb, rowstride,
                                           width, height)
                    elif self.dst_format=="BGRX" and self.src_format=="YUV444P":
                        fn_name = "J444ToARGB" if full_range else "I444ToARGB"
                        with nogil:
                            if full_range:
                                r = J444ToARGB(<const uint8_t*> y, y_stride,
                                               <const uint8_t*> u, u_stride,
                                               <const uint8_t*> v, v_stride,
                                               rgb, rowstride,
                                               width, height)
                            else:
                                r = I444ToARGB(<const uint8_t*> y, y_stride,
                                               <const uint8_t*> u, u_stride,
                                               <const uint8_t*> v, v_stride,
                                               rgb, rowstride,
                                               width, height)
                    elif self.dst_format=="RGBX" and self.src_format=="YUV420P":
                        fn_name = "J420ToABGR" if full_range else "I420ToABGR"
                        with nogil:
                            if full_range:
                                r = J420ToABGR(<const uint8_t*> y, y_stride,
                                                <const uint8_t*> u, u_stride,
                                                <const uint8_t*> v, v_stride,
                                                rgb, rowstride,
                                                width, height)
                            else:
                                r = I420ToABGR(<const uint8_t*> y, y_stride,
                                                <const uint8_t*> u, u_stride,
                                                <const uint8_t*> v, v_stride,
                                                rgb, rowstride,
                                                width, height)
                    elif self.dst_format=="BGRX" and self.src_format=="YUV420P":
                        fn_name = "J420ToARGB" if full_range else "I420ToARGB"
                        with nogil:
                            if full_range:
                                r = J420ToARGB(<const uint8_t*> y, y_stride,
                                                <const uint8_t*> u, u_stride,
                                                <const uint8_t*> v, v_stride,
                                                rgb, rowstride,
                                                width, height)
                            else:
                                r = I420ToARGB(<const uint8_t*> y, y_stride,
                                                <const uint8_t*> u, u_stride,
                                                <const uint8_t*> v, v_stride,
                                                rgb, rowstride,
                                                width, height)
                    elif self.dst_format=="RGBX" and self.src_format=="YUV444P":
                        fn_name = "J444ToABGR" if full_range else "I444ToABGR"
                        with nogil:
                            if full_range:
                                r = J444ToABGR(<const uint8_t*> y, y_stride,
                                               <const uint8_t*> u, u_stride,
                                               <const uint8_t*> v, v_stride,
                                               rgb, rowstride,
                                               width, height)
                            else:
                                r = I444ToABGR(<const uint8_t*> y, y_stride,
                                               <const uint8_t*> u, u_stride,
                                               <const uint8_t*> v, v_stride,
                                               rgb, rowstride,
                                               width, height)
                    elif self.dst_format=="BGRX" and self.src_format=="YUV444P":
                        fn_name = "J444ToARGB" if full_range else "I444ToARGB"
                        with nogil:
                            if full_range:
                                r = J444ToARGB(<const uint8_t*> y, y_stride,
                                               <const uint8_t*> u, u_stride,
                                               <const uint8_t*> v, v_stride,
                                               rgb, rowstride,
                                               width, height)
                            else:
                                r = I444ToARGB(<const uint8_t*> y, y_stride,
                                               <const uint8_t*> u, u_stride,
                                               <const uint8_t*> v, v_stride,
                                               rgb, rowstride,
                                               width, height)
        if not fn_name:
            raise RuntimeError(f"unexpected formats: src={self.src_format}, dst={self.dst_format}")
        if r!=0:
            raise RuntimeError(f"libyuv {fn_name} failed and returned {r}")
        image.free()
        cdef double elapsed = monotonic()-start
        log("libyuv.%s took %.1fms (full-range=%s)", fn_name, 1000.0*elapsed, bool(full_range))
        self.time += elapsed
        img = ImageWrapper(0, 0, self.dst_width, self.dst_height,
                            rgb_buffer, self.dst_format, Bpp*8, rowstride, Bpp, ImageWrapper.PACKED)
        img.set_full_range(full_range)
        return img

    def convert_bgrx_image(self, image: ImageWrapper) -> ImageWrapper:
        self.validate_image_src_size(image)
        cdef uint8_t *output_buffer
        cdef uint8_t *out_planes[3]
        cdef uint8_t *scaled_buffer
        cdef uint8_t *scaled_planes[3]
        cdef int i
        cdef double start = monotonic()
        cdef int iplanes = image.get_planes()
        cdef int width = image.get_width()
        cdef int height = image.get_height()
        if iplanes!=ImageWrapper.PACKED:
            raise ValueError(f"invalid plane input format: {iplanes}")
        if self.rgb_scaling:
            #first downscale:
            image = argb_scale(image, self.src_width, self.src_height, self.dst_width, self.dst_height, self.filtermode)
            width = self.dst_width
            height = self.dst_height
            self.validate_image_dst_size(image)
        cdef int stride = image.get_rowstride()
        pixels = image.get_pixels()
        assert pixels, "failed to get pixels from %s" % image
        if self.yuv_scaling:
            #re-use the same temporary buffer every time:
            output_buffer = self.output_buffer
        else:
            #allocate output buffer:
            output_buffer = <unsigned char*> memalign(self.out_buffer_size)
            if output_buffer==NULL:
                raise RuntimeError(f"failed to allocate {self.out_buffer_size} bytes for output buffer")
        for i in range(self.planes):
            #offsets are aligned, so this is safe and gives us aligned pointers:
            out_planes[i] = <uint8_t*> (memalign_ptr(<uintptr_t> output_buffer) + self.out_offsets[i])
        #get pointer to input:
        cdef int result = -1
        cdef int full_range = self.dst_full_range
        cdef const uint8_t* src
        fn_name = ""
        with buffer_context(pixels) as bc:
            src = <const uint8_t*> (<uintptr_t> int(bc))
            if self.dst_format=="NV12":
                # we can't handle full-range here!
                full_range = 0
                fn_name = "ARGBToNV12"
                with nogil:
                    result = ARGBToNV12(src, stride,
                                        out_planes[0], self.out_stride[0],
                                        out_planes[1], self.out_stride[1],
                                        width, height)
            elif self.dst_format=="YUV420P":
                if full_range:
                    fn_name = "ARGBToJ420"
                    with nogil:
                        result = ARGBToJ420(src, stride,
                                            out_planes[0], self.out_stride[0],
                                            out_planes[1], self.out_stride[1],
                                            out_planes[2], self.out_stride[2],
                                            width, height)
                else:
                    fn_name = "ARGBToI420"
                    with nogil:
                        result = ARGBToI420(src, stride,
                                            out_planes[0], self.out_stride[0],
                                            out_planes[1], self.out_stride[1],
                                            out_planes[2], self.out_stride[2],
                                            width, height)
            elif self.dst_format=="YUV444P":
                # we can't handle full-range here as there is no `ARGBToJ444`
                # we always supply full range RGB, so we can just update the metadata:
                fn_name = "ARGBToI444"
                full_range = False
                with nogil:
                    result = ARGBToI444(src, stride,
                                        out_planes[0], self.out_stride[0],
                                        out_planes[1], self.out_stride[1],
                                        out_planes[2], self.out_stride[2],
                                        width, height)
            else:
                raise RuntimeError(f"unexpected src format {self.src_format}")
        if result!=0:
            raise RuntimeError(f"libyuv.{fn_name} failed and returned {result}")
        cdef double elapsed = monotonic()-start
        log("libyuv.%s took %.1fms (full-range=%s)", fn_name, 1000.0*elapsed, bool(full_range))
        self.time += elapsed
        cdef object planes = []
        cdef object strides = []
        cdef object out_image
        if self.yuv_scaling:
            start = monotonic()
            scaled_buffer = <unsigned char*> memalign(self.scaled_buffer_size)
            if scaled_buffer==NULL:
                raise RuntimeError(f"failed to allocate {self.scaled_buffer_size} bytes for scaled buffer")
            with nogil:
                for i in range(self.planes):
                    scaled_planes[i] = scaled_buffer + self.scaled_offsets[i]
                    ScalePlane(out_planes[i], self.out_stride[i],
                               self.out_width[i], self.out_height[i],
                               scaled_planes[i], self.scaled_stride[i],
                               self.scaled_width[i], self.scaled_height[i],
                               self.filtermode)
            elapsed = monotonic()-start
            log("libyuv.ScalePlane %i times, took %.1fms", self.planes, 1000.0*elapsed)
            for i in range(self.planes):
                strides.append(self.scaled_stride[i])
                planes.append(PyMemoryView_FromMemory(<char *> scaled_planes[i], self.scaled_size[i], PyBUF_WRITE))
            self.frames += 1
            out_image = YUVImageWrapper(0, 0, self.dst_width, self.dst_height, planes, self.dst_format, 24, strides, 1, self.planes)
            out_image.cython_buffer = <uintptr_t> scaled_buffer
        else:
            #use output buffer directly:
            for i in range(self.planes):
                strides.append(self.out_stride[i])
                plane = PyMemoryView_FromMemory(<char *> out_planes[i], self.out_size[i], PyBUF_WRITE)
                planes.append(plane)
            self.frames += 1
            out_image = YUVImageWrapper(0, 0, self.dst_width, self.dst_height, planes, self.dst_format, 24, strides, 1, self.planes)
            out_image.cython_buffer = <uintptr_t> output_buffer
        if SHOW_PLANE_RANGES:
            divs = get_subsampling_divs(self.dst_format)
            for i in range(self.planes):
                xdiv, ydiv = divs[i]
                show_plane_range("YUV"[i], planes[i], self.dst_width // xdiv, strides[i], self.dst_height // ydiv)
        out_image.set_full_range(bool(full_range))
        return out_image


def selftest(full = False) -> None:
    global MAX_WIDTH, MAX_HEIGHT
    from xpra.codecs.checks import testcsc, get_csc_max_size
    from xpra.codecs.libyuv import converter
    maxw, maxh = MAX_WIDTH, MAX_HEIGHT
    testcsc(converter, full)
    if full:
        mw, mh = get_csc_max_size(converter, limit_w=32768, limit_h=32768)
        MAX_WIDTH = min(maxw, mw)
        MAX_HEIGHT = min(maxh, mh)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import tempfile

from xpra.util.env import envbool
from xpra.os_util import WIN32, OSX, POSIX
from xpra.util.io import pollwait, which
from xpra.exit_codes import exit_str
from unit.test_util import get_free_tcp_port
from unit.server_test_util import ServerTestUtil, log

TEST_RFB = envbool("XPRA_TEST_RFB", not WIN32 and not OSX)
#TEST_RFB = envbool("XPRA_TEST_RFB", False)
VFB_INITIAL_RESOLUTION = os.environ.get("XPRA_TEST_VFB_INITIAL_RESOLUTION", "1920x1080")

OPTIONS = [
    "windows",
    "notifications",
    "webcam",
    "clipboard",
    "speaker",
    "microphone",
    "av-sync",
    "printing",
    "file-transfer",
    "readonly",
    "remote-logging",
]
if not WIN32:
    OPTIONS += [
        "mmap",
    ]


class ServerMixinsOptionTestUtil(ServerTestUtil):

    @classmethod
    def setUpClass(cls):
        ServerTestUtil.setUpClass()
        cls.default_xpra_args = []
        if POSIX:
            tmpdir = tempfile.gettempdir()
            cls.default_xpra_args += [
                "--socket-dirs=%s" % tmpdir,
            ]
            if not OSX:
                cls.default_xpra_args += [
                    "--systemd-run=no",
                    "--pulseaudio=no",
                    "--start=xterm",
                ]
        cls.display = None
        cls.xvfb = None
        cls.client_display = None
        cls.client_xvfb = None

    @classmethod
    def tearDownClass(cls):
        ServerTestUtil.tearDownClass()
        if cls.xvfb:
            cls.xvfb.terminate()
            cls.xvfb = None
        if cls.client_xvfb:
            cls.client_xvfb.terminate()
            cls.client_xvfb = None

    def setUp(self):
        ServerTestUtil.setUp(self)
        if POSIX and not OSX:
            #display used by the client:
            if not self.client_display:
                self.client_display = self.find_free_display()
                self.client_xvfb = self.start_Xvfb(self.client_display)
                log("ServerMixinsOptionTest.setUpClass() client display=%s, xvfb=%s",
                    self.client_display, self.client_xvfb)
                if VFB_INITIAL_RESOLUTION:
                    xrandr = ["xrandr", "-s", VFB_INITIAL_RESOLUTION, "--display", self.client_display]
                    self.run_command(xrandr)

    def _test(self, subcommand: str, options: dict):
        log("starting test server with options=%s", options)
        # speed things up a bit by not loading: menu and codecs
        if "start-new-commands" not in options:
            options["start-new-commands"] = "no"
        for option in ("video-decoders", "video-encoders", "csc-modules"):
            if option not in options:
                options[option] = "none"
        args = ["--%s=%s" % (k, v) for k, v in options.items()]
        tcp_port = None
        xvfb = None
        if WIN32 or OSX:
            display = ""
            connect_args = []
        elif self.display:
            display = self.display
            connect_args = [display]
            args.append("--use-display=yes")
        else:
            display = self.find_free_display()
            connect_args = [display]
            if subcommand == "shadow":
                xvfb = self.start_Xvfb(display)
        if TEST_RFB or WIN32:
            tcp_port = get_free_tcp_port()
            args += ["--bind-tcp=0.0.0.0:%i" % tcp_port]
            if WIN32:
                connect_args = ["tcp://127.0.0.1:%i" % tcp_port]
        server = None
        client = None
        rfb_client = None
        gui_client = None
        try:
            log("args=%s", " ".join("'%s'" % x for x in args))
            server = self.check_server(subcommand, display, *args)
            # we should always be able to get the version:
            client = self.run_xpra(["version"] + connect_args)
            assert pollwait(client, 5) == 0, "version client failed to connect to server with args=%s" % args
            # run info query:
            cmd = ["info"] + connect_args
            client = self.run_xpra(cmd)
            r = pollwait(client, 20)
            assert r == 0, "info client failed and returned %s: '%s' for server with args=%s" % \
                           (r, exit_str(r), args)

            client_kwargs = {}
            if not (WIN32 or OSX):
                env = self.get_run_env()
                env["DISPLAY"] = self.client_display
                client_kwargs = {"env": env}

            if subcommand in ("shadow", "start-desktop") and TEST_RFB and options.get("windows", True):
                vncviewer = which("vncviewer")
                log("testing RFB clients with vncviewer '%s'", vncviewer)
                if vncviewer:
                    rfb_cmd = [vncviewer, "localhost::%i" % tcp_port]
                    rfb_client = self.run_command(rfb_cmd, **client_kwargs)
                    r = pollwait(rfb_client, 10)
                    if r is not None:
                        self.show_proc_error(rfb_client,
                                             "rfb client terminated early and returned %i for server with args=%s" % (
                                                 r, args))

            #connect a gui client:
            if WIN32 or (self.client_display and self.client_xvfb):
                xpra_args = [
                    "attach",
                    "--clipboard=no",  # could create loops
                    "--notifications=no",  # may get sent to the desktop session running the tests!
                ] + connect_args
                gui_client = self.run_xpra(xpra_args, **client_kwargs)
                r = pollwait(gui_client, 10)
                if r is not None:
                    self.show_proc_error(gui_client,
                                         "gui client terminated early and returned %i : '%s' for server with args=%s" % (
                                             r, exit_str(r), args))

            if self.display:
                self.stop_server(server, "exit", *connect_args)
            else:
                self.stop_server(server, "stop", *connect_args)
            if display:
                if display in self.dotxpra.displays():
                    log.warn("server socket for display %s should have been removed", display)

            if gui_client:
                r = pollwait(gui_client, 20)
                if r is None:
                    log.warn("client still connected!")
                    self.show_proc_pipes(server)
                    raise Exception("gui client should have been disconnected")
        except Exception:
            log.error("test error for '%s' subcommand with options=%s", subcommand, options)
            raise
        finally:
            for x in (xvfb, rfb_client, gui_client, server, client):
                try:
                    if x and x.poll() is None:
                        x.terminate()
                except OSError:
                    log("%s.terminate()", exc_info=True)

    def _test_all(self, subcommand="start"):
        #to test all:
        #TEST_VALUES = range(0, 2**len(OPTIONS)-1)
        #to test nothing disabled and everything disabled only:
        TEST_VALUES = (0, 2 ** len(OPTIONS) - 1)
        #test every option disabled individually:
        #TEST_VALUES = tuple(2**i for i in range(len(OPTIONS)))
        for i in TEST_VALUES:
            options = {}
            for o, option in enumerate(OPTIONS):
                options[option] = not bool((2 ** o) & i)
            log("test options for %i: %s", i, options)
            self._test(subcommand, options=options)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import re
import sys
import os
import glob
import shlex
from typing import Any
from collections.abc import Callable, Iterable, Sequence

from xpra.common import noop, Self
from xpra.util.str_fn import csv
from xpra.os_util import WIN32, OSX, POSIX, getuid, getgid, get_username_for_uid
from xpra.util.env import osexpand
from xpra.util.io import stderr_print, which
from xpra.util.system import is_DEB


def warn(msg: str) -> None:
    stderr_print(msg)


def is_arm() -> bool:
    import platform
    return platform.uname()[4].startswith("arm")


# can be overridden
debug = noop


class InitException(Exception):
    pass


class InitInfo(Exception):
    pass


class InitExit(Exception):
    def __init__(self, status, msg):
        self.status = status
        super().__init__(msg)


DEBUG_CONFIG_PROPERTIES: list[str] = os.environ.get("XPRA_DEBUG_CONFIG_PROPERTIES", "").split()

DEFAULT_XPRA_CONF_FILENAME: str = os.environ.get("XPRA_CONF_FILENAME", 'xpra.conf')
DEFAULT_NET_WM_NAME: str = os.environ.get("XPRA_NET_WM_NAME", "Xpra")

DEFAULT_POSTSCRIPT_PRINTER: str = ""
if POSIX:
    DEFAULT_POSTSCRIPT_PRINTER = os.environ.get("XPRA_POSTSCRIPT_PRINTER", "drv:///sample.drv/generic.ppd")
DEFAULT_PULSEAUDIO = None   # auto
if OSX or WIN32:   # pragma: no cover
    DEFAULT_PULSEAUDIO = False

# pylint: disable=import-outside-toplevel


def remove_dupes(seq: Iterable[Any]) -> list[Any]:
    seen: set[Any] = set()
    seen_add: Callable = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


_has_audio_support: bool | None = None


def has_audio_support() -> bool:
    global _has_audio_support
    if _has_audio_support is None:
        try:
            from importlib.util import find_spec
            _has_audio_support = bool(find_spec("xpra.audio"))
        except ImportError:
            _has_audio_support = False
    return bool(_has_audio_support)


def find_html5_path(page="connect.html") -> str:
    from xpra.platform.paths import get_resources_dir, get_app_dir
    return valid_html_path(
        os.path.join(get_resources_dir(), "html5", page),
        os.path.join(get_resources_dir(), "www", page),
        os.path.join(get_app_dir(), "www", page),
    )


def find_docs_path() -> str:
    from xpra.platform.paths import get_resources_dir, get_app_dir
    paths = []
    prefixes = {get_resources_dir(), get_app_dir()}
    if POSIX:
        prefixes.add("/usr/share")
        prefixes.add("/usr/local/share")
    for prefix in prefixes:
        for parts in (
                ("doc", "xpra", "index.html"),
                ("xpra", "doc", "index.html"),
                ("doc", "index.html"),
                ("doc", "index.html"),
        ):
            paths.append(os.path.join(prefix, *parts))
    return valid_html_path(*paths)


def valid_html_path(*path_options: str) -> str:
    for f in path_options:
        af = os.path.abspath(f)
        if os.path.exists(af) and os.path.isfile(af):
            return af
    return ""


def get_xorg_bin() -> str:
    xorg = os.environ.get("XPRA_XORG_BIN")
    if xorg:
        return xorg
    # Detect Xorg Binary
    if is_arm() and is_DEB() and os.path.exists("/usr/bin/Xorg"):
        # Raspbian breaks if we use a different binary..
        return "/usr/bin/Xorg"
    for p in (
        "/usr/libexec/Xorg",              # fedora 22+
        "/usr/lib/xorg/Xorg",             # ubuntu 16.10
        "/usr/lib/xorg-server/Xorg",      # arch linux
        "/usr/lib/Xorg",                  # arch linux (new 2019)
        "/usr/X11/bin/X",                 # OSX
    ):
        if os.path.exists(p):
            return p
    # look for it in $PATH:
    for x in os.environ.get("PATH", "").split(os.pathsep):   # pragma: no cover
        xorg = os.path.join(x, "Xorg")
        if os.path.isfile(xorg):
            return xorg
    return ""


def get_Xdummy_confdir() -> str:
    from xpra.platform.posix.paths import get_runtime_dir
    xrd = get_runtime_dir()
    if xrd:
        base = "${XPRA_SESSION_DIR}"
    else:   # pragma: no cover
        base = "${HOME}/.xpra"
    return base+"/xorg.conf.d/$PID"


def get_Xdummy_command(xorg_cmd="Xorg",
                       log_dir="${XPRA_SESSION_DIR}",
                       xorg_conf="${XORG_CONFIG_PREFIX}/etc/xpra/xorg.conf",
                       dpi=0) -> list[str]:
    cmd = [
        # ie: "Xorg" or "xpra_Xdummy" or "./install/bin/xpra_Xdummy"
        xorg_cmd,
        "+extension", "GLX",
        "+extension", "RANDR",
        "+extension", "RENDER",
        "-extension", "DOUBLE-BUFFER",
        "-nolisten", "tcp",
        "-noreset",
        "-novtswitch",
        "-auth", "$XAUTHORITY",
        "-logfile", f"{log_dir}/Xorg.log",
        # must be specified with some Xorg versions (ie: arch linux)
        # this directory can store xorg config files, it does not need to be created:
        "-configdir", f"{get_Xdummy_confdir()}",
        "-config", f"{xorg_conf}",
    ]
    if dpi > 0:
        cmd += ["-dpi", f"{dpi}x{dpi}"]
    return cmd


def get_Xvfb_command(width=8192, height=4096, depth=24, dpi=96) -> list[str]:
    cmd = [
        "Xvfb",
        "+extension", "GLX",
        "+extension", "Composite",
        "+extension", "RANDR",
        "+extension", "RENDER",
        "-extension", "DOUBLE-BUFFER",
        "-screen", "0", f"{width}x{height}x{depth}+32",
        # better than leaving to vfb after a resize?
        "-nolisten", "tcp",
        "-noreset",
        "-auth", "$XAUTHORITY",
    ]
    if dpi > 0:
        cmd += ["-dpi", f"{dpi}x{dpi}"]
    return cmd


def get_Xephyr_command(width=1920, height=1080, depth=24, dpi=96) -> list[str]:
    cmd = [
        "Xephyr",
        "+extension", "GLX",
        "+extension", "Composite",
        "-extension", "DOUBLE-BUFFER",
        "-screen", f"{width}x{height}x{depth}+32",
        "-nolisten", "tcp",
        "-noreset",
        "-auth", "$XAUTHORITY",
    ]
    if dpi > 0:
        cmd += ["-dpi", f"{dpi}x{dpi}"]
    return cmd


def get_weston_Xwayland_command(dpi=96) -> list[str]:
    cmd = [
        "/usr/libexec/xpra/xpra_weston_xvfb",
        "+extension", "GLX",
        "+extension", "Composite",
        "-extension", "DOUBLE-BUFFER",
        "-nolisten", "tcp",
        "-noreset",
        "-auth", "$XAUTHORITY",
    ]
    if dpi > 0:
        cmd += ["-dpi", f"{dpi}x{dpi}"]
    return cmd


def xvfb_command(cmd: str, depth=24, dpi=0) -> list[str]:
    parts = shlex.split(cmd)
    if len(parts) > 1:
        return parts
    if depth == 0:
        depth = 24
    exe = parts[0]
    if os.path.isabs(exe):
        return parts
    if exe == "Xvfb":
        return get_Xvfb_command(depth=depth, dpi=dpi)
    if exe == "Xephyr":
        return get_Xephyr_command(depth=depth, dpi=dpi)
    if exe in ("weston", "weston+Xwayland"):
        return get_weston_Xwayland_command(dpi=dpi)
    if exe in ("Xorg", "Xdummy"):
        xorg_bin = get_xorg_bin()
        return get_Xdummy_command(xorg_bin, dpi=dpi)
    if exe == "auto":
        return detect_xvfb_command(dpi=dpi)
    return parts


def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir="",
                        Xdummy_ENABLED: bool | None = None, Xdummy_wrapper_ENABLED: bool | None = None,
                        warn_fn: Callable = warn,
                        dpi=0,
                        ) -> list[str]:
    """
    This function returns the xvfb command to use.
    It can either be an `Xvfb` command or one that uses `Xdummy`,
    depending on the platform and file attributes.
    """
    if WIN32:   # pragma: no cover
        return []

    def vfb_default() -> list[str]:
        return get_Xvfb_command(dpi=dpi)

    if OSX:     # pragma: no cover
        return vfb_default()
    if sys.platform.find("bsd") >= 0 and Xdummy_ENABLED is None:  # pragma: no cover
        warn_fn(f"Warning: sorry, no support for Xdummy on {sys.platform}")
        return vfb_default()
    if is_DEB():
        # These distros do weird things and this can cause the real X11 server to crash
        # see ticket #2834
        return vfb_default()

    if Xdummy_ENABLED is False:
        return vfb_default()

    if Xdummy_ENABLED is None:
        debug("Xdummy support unspecified, will try to detect")
        # RHEL 10 needs to use `xpra_weston_xvfb` instead of Xorg:
        if os.path.exists("/etc/redhat-release"):
            with open("/etc/redhat-release") as f:
                relinfo = f.read().rstrip("\n\r")
                debug(f" found redhat-release: {relinfo!r}")
                if relinfo.find("release 10") >= 0:
                    debug(" using `xpra_weston_xvfb` on RHEL 10")
                    return get_weston_Xwayland_command(dpi)
    return detect_xdummy_command(conf_dir, bin_dir, Xdummy_wrapper_ENABLED, warn_fn, dpi=dpi)


def detect_xdummy_command(conf_dir="/etc/xpra/", bin_dir="",
                          Xdummy_wrapper_ENABLED: bool | None = None,
                          warn_fn: Callable = warn,
                          dpi=0) -> list[str]:
    if not POSIX or OSX:
        return get_Xvfb_command(dpi=dpi)
    xorg_bin = get_xorg_bin()
    if Xdummy_wrapper_ENABLED is not None:
        # honour what was specified:
        use_wrapper = Xdummy_wrapper_ENABLED
    elif not xorg_bin:
        warn_fn("Warning: Xorg binary not found, assuming the wrapper is needed!")
        use_wrapper = True
    else:
        # auto-detect
        import stat
        xorg_stat = os.stat(xorg_bin)
        if (xorg_stat.st_mode & stat.S_ISUID) != 0:
            if (xorg_stat.st_mode & stat.S_IROTH) == 0:
                warn_fn(f"{xorg_bin} is suid and not readable, Xdummy support unavailable")
                return get_Xvfb_command(dpi=dpi)
            debug(f"{xorg_bin} is suid and readable, using the xpra_Xdummy wrapper")
            use_wrapper = True
        else:
            use_wrapper = False
    xorg_conf = "${XORG_CONFIG_PREFIX}"+os.path.join(conf_dir, "xorg.conf")
    if use_wrapper:
        xorg_cmd = "xpra_Xdummy"
    else:
        xorg_cmd = xorg_bin or "Xorg"
    # so we can run from install dir:
    if bin_dir and os.path.exists(os.path.join(bin_dir, xorg_cmd)):
        if bin_dir not in os.environ.get("PATH", "/bin:/usr/bin:/usr/local/bin").split(os.pathsep):
            xorg_cmd = os.path.join(bin_dir, xorg_cmd)
    return get_Xdummy_command(xorg_cmd, xorg_conf=xorg_conf, dpi=dpi)


def wrap_cmd_str(cmd) -> str:
    xvfb_str = ""
    cr = " \\\n    "
    while cmd:
        s = ""
        while cmd:
            item = cmd[0]
            l = len(item)
            if (item.startswith("-") or item.startswith("+")) and len(cmd) > 1:
                l += len(cmd[1])
            if s and len(s)+l > 55:
                break
            v = cmd.pop(0)
            if not s:
                s += v
            else:
                s += " "+v
        if xvfb_str:
            xvfb_str += cr
        xvfb_str += s
    return xvfb_str


def get_build_info() -> list[str]:
    info = []
    try:
        from xpra.src_info import REVISION, LOCAL_MODIFICATIONS, BRANCH, COMMIT
        info.append(f"revision {REVISION}")
        if COMMIT:
            info.append(f"commit {COMMIT} from {BRANCH} branch")
        try:
            mods = int(LOCAL_MODIFICATIONS)
            info.append(f"with {mods} local changes")
        except ValueError:
            pass
    except Exception as e:
        warn(f"Error: could not find the source information: {e}")
    try:
        from xpra.build_info import build
    except ImportError as e:
        warn(f"Error: could not find the build information: {e}")
        build = {}
    info.append(build.get("type", "") + " build")
    info.insert(0, "")
    einfo = "Python " + ".".join(str(v) for v in sys.version_info[:2])
    machine = build.get("machine", "")
    if machine:
        einfo += ", "+machine
    else:
        bit = build.get("bit")
        if bit:
            einfo += ", "+bit
    info.insert(0, einfo)
    by = build.get("by", "")
    on = build.get("on", "")
    if by and on:
        info.append(f"built on {on} by {by}")
    date = build.get("date", "")
    time = build.get("time", "")
    if date and time:
        info.append(f"{date} {time}")
    cython = build.get("cython", "unknown")
    compiler = build.get("compiler", "unknown")
    if cython != "unknown" or compiler != "unknown":
        info.append("")
    if cython != "unknown":
        info.append(f"using Cython {cython}")
    if compiler != "unknown":
        cv = compiler.replace("Optimizing Compiler Version", "Optimizing Compiler\nVersion")
        info += cv.splitlines()
    return info


def name_to_field(name: str) -> str:
    return name.replace("-", "_")


def save_config(conf_file: str, config, keys, extras_types=None) -> None:
    with open(conf_file, "w", encoding="utf8") as f:
        option_types = OPTION_TYPES.copy()
        if extras_types:
            option_types.update(extras_types)
        saved = {}
        for key in keys:
            if key not in option_types:
                raise ValueError(f"invalid configuration key {key!r}")
            v = getattr(config, name_to_field(key))
            saved[key] = v
            f.write(f"{key}={v}{os.linesep}")
        debug(f"save_config: saved {saved} to {conf_file!r}")


def read_config(conf_file: str) -> dict[str, Any]:
    """
        Parses a config file into a dict of strings.
        If the same key is specified more than once,
        the value for this key will be an array of strings.
    """
    d: dict[str, str | list[str]] = {}
    if not os.path.exists(conf_file) or not os.path.isfile(conf_file):
        debug("read_config(%s) is not a file or does not exist", conf_file)
        return d
    with open(conf_file, encoding="utf8") as f:
        lines = []
        no = 0
        for line in f:
            sline = line.strip().strip('\r\n').strip()
            no += 1
            if not sline:
                debug("%4s empty line", no)
                continue
            if sline[0] in ('!', '#'):
                debug("%4s skipping comments   : %s", no, sline[:16]+"..")
                continue
            debug("%4s loaded              : %s", no, sline)
            lines.append(sline)
    debug("loaded %s lines", len(lines))
    # aggregate any lines with trailing backslash
    agg_lines = []
    l = ""
    for line in lines:
        if line.endswith("\\"):
            l += line[:-1]+" "
        else:
            l += line
            agg_lines.append(l)
            l = ""
    if l:
        # last line had a trailing backslash... meh
        agg_lines.append(l)
    debug("loaded %s aggregated lines", len(agg_lines))
    # parse name=value pairs:
    for sline in agg_lines:
        if sline.find("=") <= 0:
            debug(f"skipping line which is missing an equal sign: {sline!r}")
            continue
        props = sline.split("=", 1)
        assert len(props) == 2
        name = props[0].strip()
        value = props[1].strip()
        current_value = d.get(name)
        if current_value:
            if isinstance(current_value, list):
                d[name] = current_value + [value]
            else:
                d[name] = [current_value, value]
            debug("added to: %s='%s'", name, d[name])
        else:
            debug("assigned (new): %s='%s'", name, value)
            d[name] = value
        if name in DEBUG_CONFIG_PROPERTIES:
            from xpra.log import Logger
            log = Logger("util")
            log.info(f"{name}={d[name]} (was {current_value}), from {conf_file!r}")
    return d


def conf_files(conf_dir: str, xpra_conf_filename: str = DEFAULT_XPRA_CONF_FILENAME) -> list[str]:
    """
        Returns all the config file paths found in the config directory
        ie: ["/etc/xpra/conf.d/15_features.conf", ..., "/etc/xpra/xpra.conf"]
    """
    d: list[str] = []
    cdir = os.path.expanduser(conf_dir)
    if not os.path.exists(cdir) or not os.path.isdir(cdir):
        debug(f"invalid config directory: {cdir!r}")
        return d
    # look for conf.d subdirectory:
    conf_d_dir = os.path.join(cdir, "conf.d")
    if os.path.exists(conf_d_dir) and os.path.isdir(conf_d_dir):
        for f in sorted(os.listdir(conf_d_dir)):
            if f.endswith(".conf"):
                conf_file = os.path.join(conf_d_dir, f)
                if os.path.isfile(conf_file):
                    d.append(conf_file)
    conf_file = os.path.join(cdir, xpra_conf_filename)
    if not os.path.exists(conf_file) or not os.path.isfile(conf_file):
        debug(f"config file does not exist: {conf_file!r}")
    else:
        d.append(conf_file)
    return d


def read_xpra_conf(conf_dir: str) -> dict[str, Any]:
    """
        Reads all config files from the directory,
        returns a dict with key as strings,
        values as strings or arrays of strings.
    """
    cdir = os.path.expanduser(conf_dir)
    if not os.path.exists(cdir) or not os.path.isdir(cdir):
        debug(f"invalid config directory: {cdir!r}")
        return {}
    files = glob.glob(f"{cdir}/{DEFAULT_XPRA_CONF_FILENAME}") + glob.glob(f"{cdir}/conf.d/*.conf")
    debug(f"read_xpra_conf({conf_dir}) found conf files: {files}")
    d = {}
    for f in files:
        cd = read_config(f)
        debug(f"config({f})={cd}")
        d.update(cd)
    return d


def read_xpra_defaults(username: str | None = None, uid=None, gid=None) -> dict[str, Any]:
    """
        Reads the global <xpra_conf_filename> from the <conf_dir>
        and then the user-specific one.
        (the latter overrides values from the former)
        returns a dict with values as strings and arrays of strings.
        If the <conf_dir> is not specified, we figure out its location.
    """
    dirs = get_xpra_defaults_dirs(username, uid, gid)
    defaults: dict[str, Any] = {}
    for d in dirs:
        conf_data = read_xpra_conf(d)
        defaults.update(conf_data)
        debug(f"read_xpra_defaults: updated defaults with {d!r} : {conf_data}")
    may_create_user_config()
    return defaults


def get_xpra_defaults_dirs(username: str | None = None, uid=None, gid=None) -> list[str]:
    from xpra.platform.paths import get_default_conf_dirs, get_system_conf_dirs, get_user_conf_dirs
    # load config files in this order (the later ones override earlier ones):
    # * application defaults, ie:
    #   "/Volumes/Xpra/Xpra.app/Contents/Resources/" on OSX
    #   "C:\Program Files\Xpra\" on win32
    #   None on others
    # * system defaults, ie:
    #   "/etc/xpra" on Posix - not on OSX
    #   "/Library/Application Support/Xpra" on OSX
    #   "C:\ProgramData\Xpra" with Vista onwards
    # * user config, ie:
    #   "~/.xpra/" on all Posix, including OSX
    #   "C:\Users\<user name>\AppData\Roaming" with Visa onwards
    dirs = get_default_conf_dirs() + get_system_conf_dirs() + get_user_conf_dirs(uid)
    defaults_dirs = []
    for d in dirs:
        if not d:
            continue
        ad = osexpand(d, actual_username=username, uid=uid, gid=gid)
        if not os.path.exists(ad):
            debug(f"get_xpra_defaults_dirs: skipped missing directory {ad!r}")
            continue
        defaults_dirs.append(ad)
    return defaults_dirs


def may_create_user_config(xpra_conf_filename: str = DEFAULT_XPRA_CONF_FILENAME) -> None:
    from xpra.platform.paths import get_user_conf_dirs
    # save a user config template:
    udirs = get_user_conf_dirs()
    if any(os.path.exists(osexpand(d)) for d in udirs):
        return
    debug("no user configuration file found, trying to create one")
    for d in udirs:
        ad = os.path.expanduser(d)
        conf_file = os.path.join(ad, xpra_conf_filename)
        try:
            if not os.path.exists(ad):
                os.makedirs(ad, int('700', 8))
            with open(conf_file, 'w', encoding="utf8") as f:
                f.write("""# xpra user configuration file
# place your custom settings in this file
# they will take precedence over the system default ones.

# Examples:
# speaker=off
# dpi=144

# For more information on the file format,
# see the xpra manual at:
# https://github.com/Xpra-org/xpra/tree/master/docs/
""")
                f.flush()
            debug(f"created default config in {d!r}")
            return
        except Exception as e:
            debug(f"failed to create default config in {conf_file!r}: {e}")


OPTIONS_VALIDATION: dict[str, Callable] = {}

OPTION_TYPES: dict[str, Any] = {
    # string options:
    "encoding"          : str,
    "opengl"            : str,
    "title"             : str,
    "username"          : str,
    "password"          : str,
    "wm-name"           : str,
    "session-name"      : str,
    "dock-icon"         : str,
    "tray-icon"         : str,
    "window-icon"       : str,
    "keyboard-raw"      : bool,
    "keyboard-backend"  : str,
    "keyboard-model"    : str,
    "keyboard-layout"   : str,
    "keyboard-layouts"  : list,
    "keyboard-variant"  : str,
    "keyboard-variants" : list,
    "keyboard-options"  : str,
    "clipboard"         : str,
    "clipboard-direction" : str,
    "clipboard-filter-file" : str,
    "remote-clipboard"  : str,
    "local-clipboard"   : str,
    "pulseaudio-command": str,
    "bandwidth-limit"   : str,
    "tcp-encryption"    : str,
    "tcp-encryption-keyfile": str,
    "encryption"        : str,
    "encryption-keyfile": str,
    "pidfile"           : str,
    "mode"              : str,
    "ssh"               : str,
    "systemd-run"       : str,
    "systemd-run-args"  : str,
    "system-proxy-socket" : str,
    "chdir"             : str,
    "xvfb"              : str,
    "socket-dir"        : str,
    "sessions-dir"      : str,
    "mmap"              : str,
    "log-dir"           : str,
    "log-file"          : str,
    "border"            : str,
    "window-close"      : str,
    "min-size"          : str,
    "max-size"          : str,
    "desktop-scaling"   : str,
    "refresh-rate"      : str,
    "display"           : str,
    "download-path"     : str,
    "open-command"      : str,
    "remote-logging"    : str,
    "lpadmin"           : str,
    "lpinfo"            : str,
    "add-printer-options" : list,
    "pdf-printer"       : str,
    "postscript-printer": str,
    "debug"             : str,
    "input-method"      : str,
    "video-scaling"     : str,
    "video"             : bool,
    "audio"             : bool,
    "microphone"        : str,
    "speaker"           : str,
    "audio-source"      : str,
    "html"              : str,
    "http-scripts"      : str,
    "socket-permissions": str,
    "exec-wrapper"      : str,
    "dbus"              : bool,
    "dbus-launch"       : str,
    "webcam"            : str,
    "mousewheel"        : str,
    "input-devices"     : str,
    "shortcut-modifiers": str,
    "open-files"        : str,
    "open-url"          : str,
    "file-transfer"     : str,
    "printing"          : str,
    "headerbar"         : str,
    "challenge-handlers": list,
    # ssl options:
    "ssl"               : str,
    "ssl-key"           : str,
    "ssl-key-password"  : str,
    "ssl-cert"          : str,
    "ssl-protocol"      : str,
    "ssl-ca-certs"      : str,
    "ssl-ca-data"       : str,
    "ssl-ciphers"       : str,
    "ssl-client-verify-mode"   : str,
    "ssl-server-verify-mode"   : str,
    "ssl-verify-flags"  : str,
    "ssl-check-hostname": bool,
    "ssl-server-hostname" : str,
    "ssl-options"       : str,
    "backend"           : str,
    # int options:
    "displayfd"         : int,
    "pings"             : int,
    "quality"           : int,
    "min-quality"       : int,
    "speed"             : int,
    "min-speed"         : int,
    "compression_level" : int,
    "dpi"               : int,
    "file-size-limit"   : str,
    "idle-timeout"      : int,
    "server-idle-timeout" : int,
    "sync-xvfb"         : int,
    "pixel-depth"       : int,
    "uid"               : int,
    "gid"               : int,
    "min-port"          : int,
    "rfb-upgrade"       : int,
    "rdp-upgrade"       : bool,
    # float options:
    "auto-refresh-delay": float,
    # boolean options:
    "minimal"           : bool,
    "daemon"            : bool,
    "start-via-proxy"   : bool,
    "attach"            : bool,
    "use-display"       : str,
    "resize-display"    : str,
    "reconnect"         : bool,
    "tray"              : bool,
    "pulseaudio"        : bool,
    "mmap-group"        : str,
    "readonly"          : bool,
    "keyboard-sync"     : bool,
    "cursors"           : bool,
    "bell"              : bool,
    "notifications"     : bool,
    "xsettings"         : str,
    "system-tray"       : bool,
    "sharing"           : bool,
    "lock"              : bool,
    "delay-tray"        : bool,
    "windows"           : bool,
    "terminate-children": bool,
    "exit-with-children": bool,
    "exit-with-client"  : bool,
    "exit-with-windows" : bool,
    "exit-ssh"          : bool,
    "dbus-control"      : bool,
    "av-sync"           : bool,
    "mdns"              : bool,
    "swap-keys"         : bool,
    "commands"          : bool,
    "control"           : bool,
    "shell"             : bool,
    "http"              : bool,
    "start-new-commands": bool,
    "proxy-start-sessions": bool,
    "desktop-fullscreen": bool,
    "forward-xdg-open"  : bool,
    "modal-windows"     : bool,
    "bandwidth-detection" : bool,
    "ssl-upgrade"       : bool,
    "websocket-upgrade" : bool,
    "ssh-upgrade"       : bool,
    "splash"            : bool,
    "gstreamer"         : bool,
    "x11"               : bool,
    # arrays of strings:
    "pulseaudio-configure-commands" : list,
    "socket-dirs"       : list,
    "client-socket-dirs" : list,
    "remote-xpra"       : list,
    "encodings"         : list,
    "proxy-video-encoders" : list,
    "video-encoders"    : list,
    "csc-modules"       : list,
    "video-decoders"    : list,
    "speaker-codec"     : list,
    "microphone-codec"  : list,
    "compressors"       : list,
    "packet-encoders"   : list,
    "key-shortcut"      : list,
    "source"            : list,
    "source-start"      : list,
    "start"             : list,
    "start-late"        : list,
    "start-child"       : list,
    "start-child-late"  : list,
    "start-after-connect"       : list,
    "start-child-after-connect" : list,
    "start-on-connect"          : list,
    "start-child-on-connect"    : list,
    "start-on-last-client-exit" : list,
    "start-child-on-last-client-exit"   : list,
    "bind"              : list,
    "bind-vsock"        : list,
    "bind-tcp"          : list,
    "bind-ws"           : list,
    "bind-wss"          : list,
    "bind-ssl"          : list,
    "bind-ssh"          : list,
    "bind-rfb"          : list,
    "bind-rdp"          : list,
    "bind-quic"         : list,
    "auth"              : list,
    "vsock-auth"        : list,
    "tcp-auth"          : list,
    "ws-auth"           : list,
    "wss-auth"          : list,
    "ssl-auth"          : list,
    "ssh-auth"          : list,
    "rfb-auth"          : list,
    "rdp-auth"          : list,
    "quic-auth"         : list,
    "password-file"     : list,
    "start-env"         : list,
    "env"               : list,
}

# options removed in v6,
# don't show warnings when running with older config files:
OLD_OPTIONS: list[str] = ["fake-xinerama", "dbus-proxy"]


# in the options list, available in session files,
# but not on the command line:
NON_COMMAND_LINE_OPTIONS : list[str] = [
    "mode",
    "wm-name",
    "download-path",
    "display",
    "pdf-printer",
    "postscript-printer",
    "add-printer-options",
]

START_COMMAND_OPTIONS : list[str] = [
    "start", "start-child",
    "start-late", "start-child-late",
    "start-after-connect", "start-child-after-connect",
    "start-on-connect", "start-child-on-connect",
    "start-on-last-client-exit", "start-child-on-last-client-exit",
]
BIND_OPTIONS : list[str] = [
    "bind", "bind-tcp", "bind-ssl", "bind-ws", "bind-wss", "bind-vsock", "bind-rfb", "bind-rdp", "bind-quic",
]

# keep track of the options added since v5,
# so we can generate command lines that work with older supported versions:
OPTIONS_ADDED_SINCE_V5: list[str] = [
    "minimal", "dbus", "gstreamer",
    "keyboard-backend", "keyboard-model",
    "bind-rdp", "rdp-auth", "rdp-upgrade",
]
OPTIONS_COMPAT_NAMES: dict[str, str] = {
    "--compression_level=": "-z"
}

CLIENT_OPTIONS: list[str] = [
    "title", "username", "password", "session-name",
    "dock-icon", "tray-icon", "window-icon",
    "clipboard", "clipboard-direction", "clipboard-filter-file",
    "remote-clipboard", "local-clipboard",
    "tcp-encryption", "tcp-encryption-keyfile", "encryption", "encryption-keyfile",
    "systemd-run", "systemd-run-args",
    "socket-dir", "socket-dirs", "client-socket-dirs",
    "border", "window-close", "min-size", "max-size", "desktop-scaling",
    "file-transfer", "file-size-limit", "download-path",
    "open-command", "open-files", "printing", "open-url",
    "headerbar",
    "challenge-handlers",
    "remote-logging",
    "lpadmin", "lpinfo",
    "debug",
    "microphone", "speaker", "audio-source",
    "microphone-codec", "speaker-codec",
    "mmap", "encodings", "encoding",
    "quality", "min-quality", "speed", "min-speed",
    "compression_level",
    "dpi", "video-scaling", "auto-refresh-delay",
    "webcam", "mousewheel", "input-devices", "shortcut-modifiers", "pings",
    "tray", "keyboard-sync", "cursors", "bell", "notifications",
    "xsettings", "system-tray", "sharing", "lock",
    "delay-tray", "windows", "readonly",
    "av-sync", "swap-keys",
    "opengl",
    "start-new-commands",
    "desktop-fullscreen",
    "video-encoders", "csc-modules", "video-decoders",
    "compressors", "packet-encoders",
    "key-shortcut",
    "env",
    "ssh",
]

CLIENT_ONLY_OPTIONS: list[str] = [
    "username", "swap-keys", "dock-icon",
    "tray", "delay-tray", "tray-icon",
    "attach",
    "reconnect",
]

# options that clients can pass to the proxy
# and which will be forwarded to the new proxy instance process:
PROXY_START_OVERRIDABLE_OPTIONS: list[str] = [
    "env", "start-env", "chdir",
    "dpi",
    "encoding", "encodings",
    "quality", "min-quality", "speed", "min-speed",
    # "auto-refresh-delay",    float!
    # no corresponding command line option:
    # "wm-name", "download-path",
    "compression_level", "video-scaling",
    "title", "session-name",
    "clipboard", "clipboard-direction", "clipboard-filter-file",
    "input-method",
    "video",
    "audio", "microphone", "speaker", "audio-source", "pulseaudio",
    "idle-timeout", "server-idle-timeout",
    "use-display",
    "resize-display", "dpi", "pixel-depth",
    "readonly", "keyboard-sync", "cursors", "bell", "notifications", "xsettings",
    "system-tray", "sharing", "lock", "windows", "webcam", "html", "http-scripts",
    "terminate-children", "exit-with-children", "exit-with-client", "exit-with-windows",
    "av-sync",
    "forward-xdg-open", "modal-windows", "bandwidth-detection",
    "ssh-upgrade",
    "splash",
    "gstreamer",
    "x11",
    "printing", "file-transfer", "open-command", "open-files", "open-url", "start-new-commands",
    "mmap", "mmap-group", "mdns",
    "auth", "vsock-auth", "tcp-auth", "ws-auth", "wss-auth", "ssl-auth", "ssh-auth", "rfb-auth", "rdp-auth", "quic-auth",
    "bind", "bind-vsock", "bind-tcp", "bind-ssl", "bind-ws", "bind-wss", "bind-ssh", "bind-rfb", "bind-rdp", "bind-quic",
    "rfb-upgrade", "rdp-upgrade", "bandwidth-limit",
    "start", "start-child",
    "start-late", "start-child-late",
    "source", "source-start",
    "start-after-connect", "start-child-after-connect",
    "start-on-connect", "start-child-on-connect",
    "start-on-last-client-exit", "start-child-on-last-client-exit",
    "sessions-dir",
]
tmp: str = os.environ.get("XPRA_PROXY_START_OVERRIDABLE_OPTIONS", "")
if tmp:
    PROXY_START_OVERRIDABLE_OPTIONS = tmp.split(",")
del tmp


def get_default_key_shortcuts() -> list[str]:
    return [shortcut for e, shortcut in (
        (True, "Control+Menu:toggle_keyboard_grab"),
        (True, "Shift+Menu:toggle_pointer_grab"),
        (not OSX, "Shift+F11:toggle_fullscreen"),
        (OSX, "Control+F11:toggle_fullscreen"),
        (True, "#+F1:show_menu"),
        (True, "Control+F1:show_window_menu"),
        (True, "#+F2:show_start_new_command"),
        (True, "#+F3:show_bug_report"),
        (True, "#+F4:quit"),
        (True, "#+F5:show_window_info"),
        (True, "#+F6:show_shortcuts"),
        (True, "#+F7:show_docs"),
        (True, "#+F8:toggle_keyboard_grab"),
        (True, "#+F9:toggle_pointer_grab"),
        (True, "#+F10:magic_key"),
        (True, "#+F11:show_session_info"),
        (True, "#+F12:toggle_debug"),
        (True, "#+plus:scaleup"),
        (OSX, "#+plusminus:scaleup"),
        (True, "#+minus:scaledown"),
        (True, "#+underscore:scaledown"),
        (OSX, "#+emdash:scaledown"),
        (True, "#+KP_Add:scaleup"),
        (True, "#+KP_Subtract:scaledown"),
        (True, "#+KP_Multiply:scalereset"),
        (True, "#+bar:scalereset"),
        (True, "#+question:scalingoff"),
        (OSX, "#+degree:scalereset"),
        (OSX, "meta+grave:void"),
        (OSX, "meta+shift+asciitilde:void"),
    )
        if e]


def get_default_systemd_run() -> str:
    if WIN32 or OSX:
        return "no"
    # systemd-run was previously broken in Fedora 26:
    # https://github.com/systemd/systemd/issues/3388
    # but with newer kernels, it is working again..
    # now that we test it before using it,
    # it should be safe to leave it on auto:
    return "auto"


def get_default_pulseaudio_command() -> list[str]:
    if WIN32 or OSX:
        return []

    def description(desc):
        return f"device.description=\"{desc}\""

    def load_opt(name, **kwargs):
        args = " ".join([f"--load={name}"] + [f"{n}={v}" for n, v in kwargs.items()])
        return f"'{args}'"

    cmd = [
        "pulseaudio", "--start", "-n", "--daemonize=false", "--system=false",
        "--exit-idle-time=-1", "--load=module-suspend-on-idle",
        load_opt("module-null-sink",
                 sink_name="Xpra-Speaker",
                 sink_properties=description("Xpra\\ Speaker")),
        load_opt("module-null-sink",
                 sink_name="Xpra-Microphone",
                 sink_properties=description("Xpra\\ Microphone")),
        load_opt("module-remap-source",
                 source_name="Xpra-Mic-Source",
                 source_properties=description("Xpra\\ Mic\\ Source"),
                 master="Xpra-Microphone.monitor",
                 channels=1),
        load_opt("module-native-protocol-unix", socket="$XPRA_PULSE_SERVER"),
        load_opt("module-dbus-protocol"),
        load_opt("module-x11-publish"),
        "--log-level=2", "--log-target=stderr",
    ]
    from xpra.util.env import envbool
    if not envbool("XPRA_PULSEAUDIO_MEMFD", False):
        cmd.append("--enable-memfd=no")
    if not envbool("XPRA_PULSEAUDIO_REALTIME", True):
        cmd.append("--realtime=no")
    if not envbool("XPRA_PULSEAUDIO_HIGH_PRIORITY", True):
        cmd.append("--high-priority=no")
    return cmd


def unexpand(path: str) -> str:
    xrd = os.environ.get("XDG_RUNTIME_DIR", "")
    if POSIX and xrd and path.startswith(xrd):
        return "$XDG_RUNTIME_DIR/" + path[len(xrd):].lstrip("/")
    home = os.environ.get("HOME", "")
    if POSIX and home and path.startswith(home):
        return "~/" + path[len(home):].lstrip("/")
    return path


def unexpand_all(paths: list[str]) -> list[str]:
    return [unexpand(x) for x in paths]


GLOBAL_DEFAULTS: dict[str, Any] | None = None
# lowest common denominator here
# (the xpra.conf file shipped is generally better tuned than this - especially for 'xvfb')


def get_defaults() -> dict[str, Any]:
    global GLOBAL_DEFAULTS
    if GLOBAL_DEFAULTS is not None:
        return GLOBAL_DEFAULTS
    from xpra.platform.features import (
        OPEN_COMMAND, DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS,
        SOURCE, DEFAULT_ENV, DEFAULT_START_ENV, CAN_DAEMONIZE, SYSTEM_PROXY_SOCKET,
    )
    from xpra.platform.paths import (
        get_download_dir, get_remote_run_xpra_scripts,
        get_sessions_dir, get_socket_dirs, get_client_socket_dirs,
    )
    conf_dirs = [os.environ.get("XPRA_CONF_DIR", "")]
    build_root = os.environ.get("RPM_BUILD_ROOT")
    if build_root:
        conf_dirs.append(os.path.join(build_root, "etc", "xpra"))
    bin_dir = ""
    if sys.argv:
        xpra_cmd = sys.argv[0]
        for strip in ("/usr/bin", "/bin"):
            pos = xpra_cmd.find(strip)
            if pos >= 0:
                bin_dir = xpra_cmd[:pos+len(strip)]
                root = xpra_cmd[:pos] or "/"
                conf_dirs.append(os.path.join(root, "etc", "xpra"))
                break
    if sys.prefix == "/usr":
        conf_dirs.append("/etc/xpra")
    else:
        conf_dirs.append(os.path.join(sys.prefix, "etc", "xpra"))
    try:
        conf_dirs.append(os.getcwd())
    except FileNotFoundError:
        pass
    conf_dir = ""
    for conf_dir in conf_dirs:
        if conf_dir and os.path.exists(conf_dir):
            break
    xvfb = detect_xvfb_command(conf_dir, bin_dir, warn_fn=noop)
    xdummy = detect_xdummy_command(conf_dir, bin_dir, warn_fn=noop)

    ssl_protocol = "TLS"

    GLOBAL_DEFAULTS = {
        "encoding"          : "auto",
        "title"             : "@title@ on @hostinfo@",
        "username"          : "",
        "password"          : "",
        "wm-name"           : DEFAULT_NET_WM_NAME,
        "session-name"      : "",
        "dock-icon"         : "",
        "tray-icon"         : "",
        "window-icon"       : "",
        "keyboard-raw"      : False,
        "keyboard-backend"  : "",
        "keyboard-model"    : "",
        "keyboard-layout"   : "",
        "keyboard-layouts"  : [],
        "keyboard-variant"  : "",
        "keyboard-variants" : [],
        "keyboard-options"  : "",
        "clipboard"         : "yes",
        "clipboard-direction" : "both",
        "clipboard-filter-file" : "",
        "remote-clipboard"  : "CLIPBOARD",
        "local-clipboard"   : "CLIPBOARD",
        "pulseaudio-command": " ".join(get_default_pulseaudio_command()),
        "bandwidth-limit"   : "auto",
        "encryption"        : "",
        "tcp-encryption"    : "",
        "encryption-keyfile": "",
        "tcp-encryption-keyfile": "",
        "pidfile"           : "${XPRA_SESSION_DIR}/server.pid",
        "ssh"               : "auto",
        "systemd-run"       : get_default_systemd_run(),
        "systemd-run-args"  : "",
        "system-proxy-socket" : SYSTEM_PROXY_SOCKET,
        "xvfb"              : " ".join(xvfb),
        "xdummy"            : " ".join(xdummy),
        "chdir"             : "",
        "socket-dir"        : "",
        "sessions-dir"      : get_sessions_dir(),
        "log-dir"           : "auto",
        "log-file"          : "server.log",
        "border"            : "auto,5:off",
        "window-close"      : "auto",
        "min-size"          : "",
        "max-size"          : "",
        "desktop-scaling"   : "on",
        "refresh-rate"      : "auto",
        "display"           : "",
        "download-path"     : get_download_dir(),
        "open-command"      : " ".join(OPEN_COMMAND),
        "remote-logging"    : "both",
        "lpadmin"           : "/usr/sbin/lpadmin",
        "lpinfo"            : "/usr/sbin/lpinfo",
        "add-printer-options" : ["-u allow:$USER", "-E", "-o printer-is-shared=false"],
        "pdf-printer"       : "",
        "postscript-printer": DEFAULT_POSTSCRIPT_PRINTER,
        "debug"             : "",
        "input-method"      : "auto",
        "audio-source"      : "",
        "html"              : "auto",
        "http-scripts"      : "all",
        "socket-permissions": "600",
        "exec-wrapper"      : "",
        "dbus"              : POSIX and not OSX,
        "dbus-launch"       : "dbus-launch --sh-syntax --close-stderr",
        "webcam"            : ["auto", "no"][OSX or WIN32],
        "mousewheel"        : ["on", "invert-x"][OSX],
        "input-devices"     : "auto",
        "shortcut-modifiers": "auto",
        "open-files"        : "auto",
        "open-url"          : "auto",
        "file-transfer"     : "auto",
        "printing"          : "yes",
        "headerbar"         : ["auto", "no"][OSX or WIN32],
        "challenge-handlers": ["all"],
        # ssl options:
        "ssl"               : "auto",
        "ssl-key"           : "",
        "ssl-key-password"  : "",
        "ssl-cert"          : "",
        "ssl-protocol"      : ssl_protocol,
        "ssl-ca-certs"      : "default",
        "ssl-ca-data"       : "",
        "ssl-ciphers"       : "DEFAULT",
        "ssl-client-verify-mode"   : "optional",
        "ssl-server-verify-mode"   : "required",
        "ssl-verify-flags"  : "X509_STRICT",
        "ssl-check-hostname": True,
        "ssl-server-hostname": "",
        "ssl-options"       : "ALL,NO_COMPRESSION",
        "backend"           : "auto",
        "quality"           : 0,
        "min-quality"       : 1,
        "speed"             : 0,
        "min-speed"         : 1,
        "compression_level" : 1,
        "dpi"               : 0,
        "file-size-limit"   : "1G",
        "idle-timeout"      : 0,
        "server-idle-timeout" : 0,
        "sync-xvfb"         : None,
        "pixel-depth"       : 0,
        "uid"               : getuid(),
        "gid"               : getgid(),
        "min-port"          : 1024,
        "rfb-upgrade"       : 5,
        "rdp-upgrade"       : False,
        "auto-refresh-delay": 0.15,
        "minimal"           : False,
        "daemon"            : CAN_DAEMONIZE,
        "start-via-proxy"   : False,
        "attach"            : None,
        "use-display"       : "auto",
        "resize-display"    : ["no", "yes"][not OSX and not WIN32],
        "reconnect"         : True,
        "tray"              : True,
        "pulseaudio"        : DEFAULT_PULSEAUDIO,
        "mmap"              : "auto",
        "mmap-group"        : "auto",
        "video"             : True,
        "audio"             : True,
        "speaker"           : ["disabled", "on"][has_audio_support() and not is_arm()],
        "microphone"        : ["disabled", "off"][has_audio_support()],
        "video-scaling"     : "auto",
        "readonly"          : False,
        "keyboard-sync"     : True,
        "displayfd"         : 0,
        "pings"             : 5,
        "cursors"           : True,
        "bell"              : True,
        "notifications"     : True,
        "xsettings"         : ["auto", "no"][int(OSX or WIN32)],
        "system-tray"       : True,
        "sharing"           : None,
        "lock"              : None,
        "delay-tray"        : False,
        "windows"           : True,
        "terminate-children": False,
        "exit-with-children": False,
        "exit-with-client"  : False,
        "exit-with-windows" : False,
        "commands"          : True,
        "control"           : True,
        "shell"             : False,
        "http"              : True,
        "start-new-commands": True,
        "proxy-start-sessions": True,
        "av-sync"           : True,
        "exit-ssh"          : True,
        "dbus-control"      : not WIN32 and not OSX,
        "opengl"            : "no" if OSX else "probe",
        "mdns"              : not WIN32,
        "swap-keys"         : OSX,  # only used on osx
        "desktop-fullscreen": False,
        "forward-xdg-open"  : None,
        "modal-windows"     : False,
        "bandwidth-detection" : False,
        "ssl-upgrade"       : True,
        "websocket-upgrade" : True,
        "ssh-upgrade"       : True,
        "splash"            : None,
        "gstreamer"         : True,
        "x11"               : POSIX and not OSX,
        "pulseaudio-configure-commands"  : [" ".join(x) for x in DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS],
        "socket-dirs"       : unexpand_all(get_socket_dirs()),
        "client-socket-dirs" : unexpand_all(get_client_socket_dirs()),
        "remote-xpra"       : get_remote_run_xpra_scripts(),
        "encodings"         : ["all"],
        "proxy-video-encoders" : ["none"],
        "video-encoders"    : ["all,no-gstreamer"],
        "csc-modules"       : ["all"],
        "video-decoders"    : ["all,no-gstreamer"],
        "speaker-codec"     : [],
        "microphone-codec"  : [],
        "compressors"       : ["all"],
        "packet-encoders"   : ["all"],
        "key-shortcut"      : get_default_key_shortcuts(),
        "bind"              : ["auto"],
        "bind-vsock"        : [],
        "bind-tcp"          : [],
        "bind-ws"           : [],
        "bind-wss"          : [],
        "bind-ssl"          : [],
        "bind-ssh"          : [],
        "bind-rfb"          : [],
        "bind-rdp"          : [],
        "bind-quic"         : [],
        "auth"              : [],
        "vsock-auth"        : [],
        "tcp-auth"          : [],
        "ws-auth"           : [],
        "wss-auth"          : [],
        "ssl-auth"          : [],
        "ssh-auth"          : [],
        "rfb-auth"          : [],
        "rdp-auth"          : [],
        "quic-auth"         : [],
        "password-file"     : [],
        "source"            : SOURCE,
        "source-start"      : [],
        "start"             : [],
        "start-late"        : [],
        "start-child"       : [],
        "start-child-late"  : [],
        "start-after-connect"       : [],
        "start-child-after-connect" : [],
        "start-on-connect"          : [],
        "start-child-on-connect"    : [],
        "start-on-last-client-exit" : [],
        "start-child-on-last-client-exit"   : [],
        "start-env"         : list(DEFAULT_ENV),
        "env"               : list(DEFAULT_START_ENV),
    }
    return GLOBAL_DEFAULTS


# fields that got renamed:
CLONES: dict[str, str] = {}

# these options should not be specified in config files:
NO_FILE_OPTIONS = ("daemon", )


TRUE_OPTIONS: Sequence[str | bool] = ("yes", "true", "1", "on", True)
FALSE_OPTIONS: Sequence[str | bool] = ("no", "false", "0", "off", False)
ALL_BOOLEAN_OPTIONS: Sequence[str | bool] = tuple(list(TRUE_OPTIONS)+list(FALSE_OPTIONS))
OFF_OPTIONS: Sequence[str] = ("off", )


def str_to_bool(v: Any, default: bool = True) -> bool:
    if isinstance(v, str):
        v = v.lower().strip()
    if v in TRUE_OPTIONS:
        return True
    if v in FALSE_OPTIONS:
        return False
    return default


def parse_bool_or(k: str, v: Any, auto: bool | None = None) -> bool | None:
    if isinstance(v, str):
        v = v.lower().strip()
    if v in TRUE_OPTIONS:
        return True
    if v in FALSE_OPTIONS:
        return False
    if v in ("auto", None):
        # keep default - which may be None!
        return auto
    try:
        return bool(int(v))
    except ValueError:
        warn(f"Warning: cannot parse value {v!r} for {k!r} as a boolean")
        return auto


def print_bool(k, v, true_str='yes', false_str='no') -> str:
    if v is None:
        return "auto"
    if isinstance(v, bool):
        if v:
            return true_str
        return false_str
    warn(f"Warning: cannot print value {v!r} for {k!r} as a boolean")
    return ""


def parse_bool_or_int(k, v) -> int | float | bool:
    return parse_bool_or_number(int, k, v)


def parse_bool_or_number(numtype:Callable, k:str, v, auto=0) -> int | float | bool:
    if isinstance(v, str):
        v = v.lower()
    if v in TRUE_OPTIONS:
        return 1
    if v in FALSE_OPTIONS:
        return 0
    return parse_number(numtype, k, v, auto)


def parse_number(numtype, k, v, auto=0) -> int | float:
    if isinstance(v, str):
        v = v.lower()
    if v == "auto":
        return auto
    try:
        return numtype(v)
    except (ValueError, TypeError) as e:
        warn(f"Warning: cannot parse value {v!r} for {k} as a type {numtype}: {e}")
        return auto


def print_number(i, auto_value=0) -> str:
    if i == auto_value:
        return "auto"
    return str(i)


def parse_with_unit(numtype:str, v, subunit="bps", min_value=250000) -> int | None:
    if isinstance(v, int):
        return v
    # special case for bandwidth-limit, which can be specified using units:
    try:
        v = str(v).lower().strip()
        if not v or v in FALSE_OPTIONS:
            return 0
        if v == "auto":
            return None
        r = re.match(r'([0-9.]*)(.*)', v)
        assert r
        f = float(r.group(1))
        unit = r.group(2).lower().strip()
        if unit.endswith(subunit):
            unit = unit[:-len(subunit)]     # ie: 10mbps -> 10m
        if unit == "k":
            f *= 1000
        elif unit == "m":
            f *= 1000000
        elif unit == "g":
            f *= 1000000000
        elif unit in ("", "b"):
            pass    # no multiplier
        else:
            raise ValueError(f"unknown unit {unit!r}")
        if min_value is not None and f < min_value:
            raise ValueError(f"value {f} is too low, minimum is {min_value}")
        return int(f)
    except Exception as e:
        raise InitException(f"invalid value for {numtype} {v!r}: {e}") from None


def validate_config(d=None, discard=NO_FILE_OPTIONS, extras_types=None, extras_validation=None) -> dict[str,Any]:
    return do_validate_config(d or {}, discard, extras_types or {}, extras_validation or {})


def do_validate_config(d:dict, discard, extras_types:dict, extras_validation:dict) -> dict[str, Any]:
    """
        Validates all the options given in a dict with fields as keys and
        strings or arrays of strings as values.
        Each option is strongly typed and invalid value are discarded.
        We get the required datatype from OPTION_TYPES
    """
    validations = OPTIONS_VALIDATION.copy()
    validations.update(extras_validation)
    option_types = OPTION_TYPES.copy()
    option_types.update(extras_types)
    nd: dict[str, Any] = {}
    for k, v in d.items():
        if k in discard:
            warn(f"Warning: option {k!r} is not allowed in configuration files")
            continue
        vt = option_types.get(k)
        if vt is None:
            if k in OLD_OPTIONS:
                continue
            warn(f"Warning: invalid option: {k!r}")
            continue
        if vt is str:
            if not isinstance(v, str):
                warn(f"invalid value for {k!r}: {type(v)} (string required)")
                continue
        elif vt is int:
            v = parse_bool_or_number(int, k, v)
            if v is None:
                continue
        elif vt is float:
            v = parse_number(float, k, v)
            if v is None:
                continue
        elif vt is bool:
            v = parse_bool_or(k, v)
            if v is None:
                continue
        elif vt is list:
            if isinstance(v, str):
                # could just be that we specified it only once..
                v = [v]
            elif isinstance(v, list) or v is None:
                # ok so far..
                pass
            else:
                warn(f"Warning: invalid value for {k!r}: {type(v)} (a string or list of strings is required)")
                continue
        else:
            warn(f"Error: unknown option type for {k!r}: {vt}")
        validation = validations.get(k)
        if validation and v is not None:
            msg = validation(v)
            if msg:
                warn(f"Warning: invalid value for {k!r}: {v}, {msg}")
                continue
        nd[k] = v
    return nd


class XpraConfig:
    def __repr__(self):
        return f"XpraConfig({self.__dict__})"

    def clone(self) -> Self:
        c = XpraConfig()
        c.__dict__ = dict(self.__dict__)
        return c


def make_defaults_struct(extras_defaults=None, extras_types=None, extras_validation=None,
                         username="", uid=getuid(), gid=getgid()) -> XpraConfig:
    return do_make_defaults_struct(extras_defaults or {}, extras_types or {}, extras_validation or {},
                                   username, uid, gid)


def do_make_defaults_struct(extras_defaults: dict, extras_types: dict, extras_validation: dict,
                            username:str, uid: int, gid: int) -> XpraConfig:
    # populate config with default values:
    if not username and uid:
        username = get_username_for_uid(uid)
    defaults = read_xpra_defaults(username, uid, gid)
    return dict_to_validated_config(defaults, extras_defaults, extras_types, extras_validation)


def dict_to_validated_config(d: dict, extras_defaults=None, extras_types=None, extras_validation=None) -> XpraConfig:
    options = get_defaults().copy()
    if extras_defaults:
        options.update(extras_defaults)
    # parse config:
    validated = validate_config(d, extras_types=extras_types, extras_validation=extras_validation)
    options.update(validated)
    for k,v in CLONES.items():
        if k in options:
            options[v] = options[k]
    return dict_to_config(options)


def dict_to_config(options) -> XpraConfig:
    config = XpraConfig()
    for k,v in options.items():
        setattr(config, name_to_field(k), v)
    return config


def fixup_debug_option(value: str) -> str:
    """ backwards compatible parsing of the debug option, which used to be a boolean """
    if not value:
        return ""
    value = str(value)
    if value.strip().lower() in ("yes", "true", "on", "1"):
        return "all"
    if value.strip().lower() in ("no", "false", "off", "0"):
        return ""
    # if we're here, the value should be a CSV list of categories
    return value


def csvstr(value) -> str:
    if isinstance(value, (list, tuple, set, dict)):
        return ",".join(str(x).strip() for x in value if x)
    if isinstance(value, str):
        return value.strip()
    raise ValueError(f"don't know how to convert {type(value)} to a csv list!")


def csvstrl(value) -> str:
    return csvstr(value).lower()


def nodupes(s) -> list[str]:
    return remove_dupes(x.strip().lower() for x in s.split(","))


def fixup_socketdirs(options) -> None:
    for option_name in ("socket_dirs", "client_socket_dirs"):
        value = getattr(options, option_name)
        if isinstance(value, str):
            value = value.split(os.path.pathsep)
        else:
            assert isinstance(getattr(options, option_name), (list, tuple))
            value = [v for x in value for v in x.split(os.path.pathsep)]
        setattr(options, option_name, value)


def fixup_pings(options) -> None:
    # pings used to be a boolean, True mapped to "5"
    if isinstance(options.pings, int):
        return
    try:
        pings = str(options.pings).lower()
        if pings in TRUE_OPTIONS:
            options.pings = 5
        elif pings in FALSE_OPTIONS:
            options.pings = 0
        else:
            options.pings = int(options.pings)
    except ValueError:
        options.pings = 5


def fixup_encodings(options) -> None:
    from xpra.codecs.constants import PREFERRED_ENCODING_ORDER
    estr = csvstr(options.encodings)
    RENAME = {
        "jpg"   : "jpeg",
        "png/l" : "png/L",
        "pngl"  : "png/L",
        "png/p" : "png/P",
        "pngp"  : "png/P",
    }
    options.encoding = RENAME.get(options.encoding, options.encoding)
    encodings = [RENAME.get(x, x) for x in nodupes(estr)]
    while True:
        try:
            i = encodings.index("all")
        except ValueError:
            break
        else:
            # replace 'all' with the actual value:
            encodings = encodings[:i]+list(PREFERRED_ENCODING_ORDER)+encodings[i+1:]
    # if the list only has items to exclude (ie: '-scroll,-jpeg')
    # then 'all' is implied:

    def isneg(enc: str) -> bool:
        return enc.startswith("-") or enc.startswith("no-")

    def stripneg(enc) -> str:
        if enc.startswith("-"):
            return enc[1:]
        if enc.startswith("no-"):
            return enc[3:]
        return enc

    if not any(True for e in encodings if not isneg(e)):
        encodings = list(PREFERRED_ENCODING_ORDER)+encodings
    if "rgb" in encodings:
        if "rgb24" not in encodings:
            encodings.append("rgb24")
        if "rgb32" not in encodings:
            encodings.append("rgb32")
    encodings = remove_dupes(encodings)
    invalid = [stripneg(e) for e in encodings if stripneg(e) not in PREFERRED_ENCODING_ORDER]
    if invalid:
        from xpra.exit_codes import ExitCode
        raise InitExit(ExitCode.UNSUPPORTED, "invalid encodings specified: " + csv(invalid))
    # remove the negated encodings:
    for rm in tuple(stripneg(e) for e in encodings if isneg(e)):
        while True:
            try:
                encodings.remove(rm)
            except ValueError:
                break
    options.encodings = encodings
    if not options.video:
        options.csc_modules = ["none"]
        options.video_encoders = ["none"]
        options.video_decoders = ["none"]
    else:
        from xpra.codecs.video import ALL_VIDEO_ENCODER_OPTIONS, ALL_CSC_MODULE_OPTIONS, ALL_VIDEO_DECODER_OPTIONS
        for name, all_list in (
            ("csc-modules", ALL_CSC_MODULE_OPTIONS),
            ("video-encoders", ALL_VIDEO_ENCODER_OPTIONS),
            ("video-decoders", ALL_VIDEO_DECODER_OPTIONS),
        ):
            # ensure value is always a sequence:
            attr_name = name.replace("-", "_")
            value = getattr(options, attr_name)
            if not isinstance(value, Sequence):
                value = [x for x in csvstrl(value).split(",") if x.strip()]
                setattr(options, attr_name, value)
            if "help" in value:
                raise InitInfo(f"the following {name} are defined: {csv(all_list)}")


def fixup_compression(options) -> None:
    from xpra.net import compression
    cstr = csvstrl(options.compressors)
    if cstr == "all":
        compressors = compression.PERFORMANCE_ORDER
    elif cstr == "none":
        compressors = ()
    else:
        compressors = nodupes(cstr)
        unknown = tuple(x for x in compressors if x and x not in compression.ALL_COMPRESSORS)
        if unknown:
            warn("Warning: invalid compressor(s) specified: " + csv(unknown))
        # keep only valid ones
        # (ignores `lzo` and `zlib` which have been removed but may still exist in config files)
        compressors = [x for x in compressors if x in compression.VALID_COMPRESSORS]
    options.compressors = list(compressors)


def fixup_packetencoding(options) -> None:
    from xpra.net import packet_encoding
    pestr = csvstrl(options.packet_encoders)
    if pestr == "all":
        packet_encoders = packet_encoding.PERFORMANCE_ORDER
    else:
        packet_encoders = nodupes(pestr)
        unknown = [x for x in packet_encoders if x and x not in packet_encoding.ALL_ENCODERS]
        if unknown:
            warn("Warning: invalid packet encoder(s) specified: " + csv(unknown))
        # keep only valid ones
        # (ignores `rencode`, `bencode` and `yaml` which have been removed but may still exist in config files)
        packet_encoders = [x for x in packet_encoders if x in packet_encoding.VALID_ENCODERS]
    options.packet_encoders = packet_encoders


def fixup_keyboard(options) -> None:
    # variants and layouts can be specified as CSV, convert them to lists:
    def p(v) -> list[str]:
        try:
            if isinstance(v, Sequence):
                seq = v
            else:
                seq = str(v).split(",")
            r = remove_dupes(x.strip() for x in seq)
            # remove empty string if that's the only value:
            if r and len(r) == 1 and r[0] == "":
                r = []
            return r
        except Exception:
            return []

    options.keyboard_backend = "" if options.keyboard_backend == "auto" else options.keyboard_backend
    options.keyboard_layouts = p(options.keyboard_layouts)
    options.keyboard_variants = p(options.keyboard_variants)
    options.keyboard_raw = str_to_bool(options.keyboard_raw)


def fixup_clipboard(options) -> None:
    cd = options.clipboard_direction.lower().replace("-", "")
    if cd == "toserver":
        options.clipboard_direction = "to-server"
    elif cd == "toclient":
        options.clipboard_direction = "to-client"
    elif cd == "both":
        options.clipboard_direction = "both"
    elif cd in ("disabled", "none"):
        options.clipboard_direction = "disabled"
    else:
        warn(f"Warning: invalid value for clipboard-direction: {options.clipboard_direction!r}")
        warn(" specify 'to-server', 'to-client' or 'both'")
        options.clipboard_direction = "disabled"


def abs_paths(options) -> None:
    ew = options.exec_wrapper
    if ew:
        ewp = shlex.split(ew)
        if ewp and not os.path.isabs(ewp[0]):
            abscmd = which(ewp[0])
            if abscmd:
                ewp[0] = abscmd
                options.exec_wrapper = shlex.join(ewp)
    # convert to absolute paths before we daemonize
    for k in ("clipboard-filter-file",
              "tcp-encryption-keyfile", "encryption-keyfile",
              "log-dir",
              "download-path", "exec-wrapper",
              "ssl-key", "ssl-cert", "ssl-ca-certs"):
        f = k.replace("-", "_")
        v = getattr(options, f)
        if v and (k != "ssl-ca-certs" or v != "default"):
            if os.path.isabs(v) or v == "auto":
                continue
            if v.startswith("~") or v.startswith("$"):
                continue
            setattr(options, f, os.path.abspath(v))


def fixup_options(options) -> None:
    fixup_encodings(options)
    fixup_pings(options)
    fixup_compression(options)
    fixup_packetencoding(options)
    fixup_socketdirs(options)
    fixup_clipboard(options)
    fixup_keyboard(options)
    abs_paths(options)
    # remote-xpra is meant to be a list, but the user can specify a string using the command line,
    # in which case we replace all the default values with this single entry:
    if not isinstance(options.remote_xpra, (list, tuple)):
        options.remote_xpra = [options.remote_xpra]


def main(argv):
    from xpra.util.str_fn import nonl

    def print_options(o):
        for k, ot in sorted(OPTION_TYPES.items()):
            v = getattr(o, name_to_field(k), "")
            if ot is bool and v is None:
                v = "Auto"
            if isinstance(v, list):
                v = csv(str(x) for x in v)
            print("* %-32s : %s" % (k, nonl(v)))
    from xpra.platform import program_context
    from xpra.log import enable_color, consume_verbose_argv
    with program_context("Config-Info", "Config Info"):
        enable_color()
        if consume_verbose_argv(argv, "all"):
            global debug

            def print_debug(*args):
                print(args[0] % args[1:])
            debug = print_debug

        if len(argv) > 1:
            for filename in argv[1:]:
                print("")
                print(f"Configuration file {filename!r}")
                if not os.path.exists(filename):
                    print(" Error: file not found")
                    continue
                d = read_config(filename)
                config = dict_to_validated_config(d)
                print_options(config)
        else:
            print("Default Configuration:")
            print_options(make_defaults_struct())


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3

# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=bare-except

import datetime
from shutil import which
from subprocess import Popen, PIPE, STDOUT, getstatusoutput, run
from typing import Any
from collections.abc import Sequence
import socket
import platform
import os.path
import sys


SRC_INFO_FILE = "xpra/src_info.py"
BUILD_INFO_FILE = "xpra/build_info.py"


def update_properties(props: dict[str, Any], filename: str) -> None:
    eprops = get_properties(filename)
    eprops.update(props)
    save_properties(eprops, filename)


def save_properties(props: dict[str, Any], filename: str) -> None:
    if os.path.exists(filename):
        try:
            os.unlink(filename)
        except OSError:
            print(f"WARNING: failed to delete {filename!r}")
    print(f"# updated {filename!r} with:")
    with open(filename, mode="w") as f:
        for k in sorted(props.keys()):
            v = props[k]
            pair = f"{k} = {v!r}"
            f.write(f"{pair}\n")
            print(pair)


def get_properties(filename: str) -> dict[str, Any]:
    props = dict()
    if not os.path.exists(filename):
        return props

    from importlib.util import spec_from_file_location, module_from_spec
    spec = spec_from_file_location("xpra", filename)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)

    for attr in dir(module):
        if attr.startswith("_"):
            continue
        props[attr] = getattr(module, attr)
    return props


def get_machineinfo() -> str:
    if platform.uname()[4]:
        return platform.uname()[4]
    return "unknown"


def get_cpuinfo() -> str:
    if platform.uname()[5]:
        return platform.uname()[5]
    if os.path.exists("/proc/cpuinfo"):
        with open("/proc/cpuinfo", "r") as f:
            for line in f:
                if line.startswith("model name"):
                    return line.split(": ")[1].replace("\n", "").replace("\r", "")
    return "unknown"


def get_status_output(*args, **kwargs) -> tuple[int, bytes, bytes]:
    kwargs["stdout"] = PIPE
    kwargs["stderr"] = STDOUT
    try:
        p = Popen(*args, **kwargs)
    except Exception as e:
        print("error running %s,%s: %s" % (args, kwargs, e))
        return -1, "", ""
    stdout, stderr = p.communicate()
    return p.returncode, stdout, stderr


def get_output_lines(cmd, valid_exit_code=0) -> Sequence[str]:
    try:
        returncode, stdout, stderr = get_status_output(cmd, shell=True)
        if returncode != valid_exit_code:
            print("'%s' failed with return code %s" % (cmd, returncode))
            print("stderr: %s" % stderr)
            return ()
        if not stdout:
            print("could not get output from command '%s'" % (cmd,))
            return ()
        out = stdout.decode('utf-8')
        return out.splitlines()
    except Exception as e:
        print("error running '%s': %s" % (cmd, e))
    return ()


def get_first_line_output(cmd, valid_exit_code=0) -> str:
    lines = get_output_lines(cmd, valid_exit_code)
    if lines:
        return lines[0]
    return ""


def get_nvcc_version() -> str:
    if sys.platform == "darwin":
        return ""
    for nvcc in ("/usr/local/cuda/bin/nvcc", "/opt/cuda/bin/nvcc", which("nvcc")):
        if nvcc and os.path.exists(nvcc):
            cmd = f"{nvcc} --version"
            lines = get_output_lines(cmd)
            if lines:
                vline = lines[-1]
                vpos = vline.rfind(", V")
                if vpos>0:
                    return vline[vpos+3:]
    return ""


def get_compiler_version() -> str:
    cc_version = "%s --version" % os.environ.get("CC", "gcc")
    if sys.platform == "darwin":
        lines = get_output_lines(cc_version)
        for line in lines:
            if line.startswith("Apple"):
                return line
        return ""
    return get_first_line_output(cc_version)


def get_linker_version() -> str:
    if sys.platform == "darwin":
        ld_version = "%s -v" % os.environ.get("LD", "ld")
        lines = get_output_lines(ld_version)
        for line in lines:
            if line.find("using: ")>0:
                return line.split("using: ", 1)[1]
        return ""
    ld_version = "%s --version" % os.environ.get("LD", "ld")
    return get_first_line_output(ld_version)


def get_platform_name() -> str:
    # better version info than standard python platform:
    if sys.platform.startswith("sun"):
        # couldn't find a better way to distinguish opensolaris from solaris...
        with open("/etc/release", "r", encoding="latin1") as f:
            data = f.read()
        if data and str(data).lower().find("opensolaris"):
            return "OpenSolaris"
        return "Solaris"
    if sys.platform == "darwin":
        try:
            # ie: MacOS 10.14.6
            return "MacOS %s" % platform.mac_ver()[0]
        except (AttributeError, TypeError, IndexError):
            return "MacOS"
    if sys.platform.find("openbsd")>=0:
        return "OpenBSD"
    if sys.platform.startswith("win"):
        try:
            out = run(
                [
                    "powershell",
                    "-NoProfile",
                    "-Command",
                    "$OutputEncoding = [Console]::OutputEncoding = New-Object System.Text.UTF8Encoding; (Get-CimInstance Win32_OperatingSystem).Caption | Out-String",
                 ],
                 capture_output=True,
                 text=True,
                 encoding="utf-8",
            ).stdout.strip()
            return out
        except OSError:
            pass
        return "Microsoft Windows"
    if sys.platform.find("bsd") >= 0:
        return "BSD"
    try:
        from xpra.util.system import get_linux_distribution
        ld = get_linux_distribution()
        ldvalid = tuple(x for x in ld if x not in ("", "unknown", "n/a"))
        if ldvalid:
            return "Linux %s" % (" ".join(ldvalid))
    except Exception:
        pass
    return sys.platform


def get_build_props() -> dict[str, Any]:
    props = {}
    source_epoch = os.environ.get("SOURCE_DATE_EPOCH")
    if source_epoch:
        # reproducible builds:
        build_time = datetime.datetime.fromtimestamp(int(source_epoch), tz=datetime.timezone.utc)
        build_date = build_time.date()
    else:
        # win32, macos and older build environments:
        build_time = datetime.datetime.now()
        build_date = datetime.date.today()
        # also record username and hostname:
        try:
            import getpass
            props["user"] = getpass.getuser()
        except OSError:
            props["user"] = os.environ.get("USER")
        props["on"] = socket.gethostname()

    props.update({
        "date": build_date.isoformat(),
        "time": build_time.strftime("%H:%M"),
        "machine": get_machineinfo(),
        "cpu": get_cpuinfo(),
        "bit": platform.architecture()[0],
        "os": get_platform_name(),
        "type": os.environ.get("BUILD_TYPE", ""),
    })
    try:
        from Cython import __version__ as cython_version
    except ImportError:
        cython_version = "unknown"
    props.update({
        "python": ".".join(str(x) for x in sys.version_info[:3]),
        "cython": cython_version,
        "compiler": get_compiler_version(),
        "nvcc": get_nvcc_version(),
        "linker": get_linker_version(),
    })
    return props


def get_libs() -> dict[str, Any]:
    # record pkg-config versions:
    PKG_CONFIG = os.environ.get("PKG_CONFIG", "pkg-config")
    libs: dict[str, Any] = {}
    if os.name == "nt":
        returncode, out, _ = get_status_output(["pacman", "-Q"])
        if returncode == 0:
            for line in out.decode().splitlines():
                parts = line.split(" ")
                if len(parts) != 2:
                    continue
                pkg_name, version = parts
                libs[pkg_name] = version
    elif sys.platform == "darwin":
        returncode, out, _ = get_status_output(["jhbuild", "list", "-a", "-r"])
        if returncode == 0:
            for line in out.decode().splitlines():
                parts = line.split(" ")
                if len(parts) != 2:
                    continue
                pkg_name, version = parts
                if pkg_name == "Modules":
                    continue
                libs[pkg_name] = version.lstrip("(").rstrip(")")
    else:
        for pkg in (
            "libc",
            "vpx", "x264", "webp", "yuv", "nvenc", "nvfbc",
            "nvenc",
            "x11", "xrandr", "xtst", "xfixes", "xkbfile", "xcomposite", "xdamage", "xext",
            "gobject-introspection-1.0",
            "gtk+-3.0", "py3cairo", "pygobject-3.0", "gtk+-x11-3.0",
            "python3",
        ):
            # fugly magic for turning the package atom into a legal variable name:
            cmd = [PKG_CONFIG, "--modversion", pkg]
            returncode, out, _ = get_status_output(cmd)
            if returncode == 0:
                libs[pkg] = out.decode().replace("\n", "").replace("\r", "")
    return libs


def get_python_libs() -> dict[str, Any]:
    # we could potentially limit this data collection to win32 and macos?
    python_libs: dict[str, Any] = {}
    returncode, out, _ = get_status_output(["pip3", "freeze"])
    if returncode == 0:
        python_libs: dict[str, Any] = {}
        for line in out.decode().splitlines():
            parts = line.split("==")
            if len(parts) != 2:
                continue
            pkg_name, version = parts
            python_libs[pkg_name] = version
    return python_libs


def record_build_info() -> None:
    props = get_properties(BUILD_INFO_FILE)
    props.update({
        "build": get_build_props(),
    })
    if os.environ.get("RECORD_LIBS", "1") == "1":
        props.update({
            "libs": get_libs(),
            "python_libs": get_python_libs(),
        })
    if sys.platform == "darwin":
        sbom = {}
        packages = {}
        SKIPPED = ("xar", "cpio", "bomutils", )
        for package_name in get_jhbuild_package_list():
            if package_name in SKIPPED or package_name.startswith("meta-"):
                continue
            pinfo = get_jhbuild_package_info(package_name)
            if pinfo:
                sbom[package_name] = (0, "", package_name, pinfo.get("Version", ""))
                packages[package_name] = {key: value for key, value in pinfo.items() if key in ("Name", "Version", "URL")}
        props["sbom"] = sbom
        props["packages"] = packages
    save_properties(props, BUILD_INFO_FILE)


def get_jhbuild_package_list() -> list[str]:
    cmd = "jhbuild list"
    r, output = getstatusoutput(cmd)
    if r:
        print(f"`jhbuild list` failed and returned {r}")
        return []
    packages = []
    for line in output.split("\n"):
        if line.find(" ") >= 0:
            continue
        packages.append(line.strip())
    return packages


def get_jhbuild_package_info(name: str) -> dict[str, str]:
    cmd = f"jhbuild info {name}"
    r, output = getstatusoutput(cmd)
    if r:
        print(f"`jhbuild info {name}` failed and returned {r}")
        return {}
    props: dict[str, str] = {}
    for line in output.split("\n"):
        parts = line.split(":", 1)
        if len(parts) == 2:
            props[parts[0].strip()] = parts[1].strip()
    return props


def get_vcs_props():
    props = {
        "REVISION": "unknown",
        "LOCAL_MODIFICATIONS": 0,
        "BRANCH": "unknown",
        "COMMIT": "unknown"
    }
    branch = None
    for cmd in (
        r"git branch --show-current",
        # when in detached state, the one above does not work, but this one does:
        r"git branch --remote --verbose --no-abbrev --contains | sed -rne 's/^[^\/]*\/([^\ ]+).*$/\1/p'",
        # if all else fails:
        r"git branch | grep '* '",
    ):
        proc = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
        out, _ = proc.communicate()
        if proc.returncode == 0:
            branch_out = out.decode("utf-8").splitlines()
            if branch_out:
                branch = branch_out[0]
                if branch.startswith("* "):
                    branch = branch[2:]
                break
    if not branch:
        print("Warning: could not get branch information")
    else:
        props["BRANCH"] = branch

    # use the number of changes since the last tag:
    proc = Popen("git describe --long --always --tags", stdout=PIPE, stderr=PIPE, shell=True)
    out, _ = proc.communicate()
    if proc.returncode != 0:
        print("'git describe --long --always --tags' failed with return code %s" % proc.returncode)
        return props
    if not out:
        print("could not get version information")
        return props
    out = out.decode('utf-8').splitlines()[0]
    # ie: out=v4.0.6-58-g6e6614571
    parts = out.split("-")
    if len(parts) == 1:
        commit = parts[0]
        print("could not get revision number - no tags?")
        rev_str = "0"
    elif len(parts) == 3:
        rev_str = parts[1]
        commit = parts[2]
    else:
        print("could not parse version information from string: %s" % out)
        return props
    props["COMMIT"] = commit
    try:
        rev = int(rev_str)
    except ValueError:
        print("could not parse revision counter from string: %s (original version string: %s)" % (rev_str, out))
        return props

    if branch == "master":
        # fake a sequential revision number that continues where svn left off,
        # by counting the commits and adding a magic value (5014)
        proc = Popen("git rev-list --count HEAD --first-parent", stdout=PIPE, stderr=PIPE, shell=True)
        out, _ = proc.communicate()
        if proc.returncode != 0:
            print("failed to get commit count using 'git rev-list --count HEAD'")
            sys.exit(1)
        rev = int(out.decode("utf-8").splitlines()[0]) + 5014
    props["REVISION"] = rev
    # find number of local files modified:
    changes = 0
    proc = Popen("git status", stdout=PIPE, stderr=PIPE, shell=True)
    (out, _) = proc.communicate()
    if proc.poll() != 0:
        print("could not get status of local files")
        return props

    lines = out.decode('utf-8').splitlines()
    for line in lines:
        sline = line.strip()
        if sline.startswith("modified: ") or sline.startswith("new file:") or sline.startswith("deleted:"):
            changes += 1
    props["LOCAL_MODIFICATIONS"] = changes
    return props


def record_src_info() -> None:
    update_properties(get_vcs_props(), SRC_INFO_FILE)


def check_file(filename: str) -> bool:
    return os.path.exists(filename) and os.path.isfile(filename) and os.stat(filename).st_size > 0


def main(args):
    if not check_file(SRC_INFO_FILE) or "src" in args:
        record_src_info()
    if not check_file(BUILD_INFO_FILE) or "build" in args:
        record_build_info()
    if "revision" in args:
        props = get_vcs_props()
        try:
            mods = int(props.get("LOCAL_MODIFICATIONS"))
        except ValueError:
            mods = 0
        commit = props.get("COMMIT")
        print("%s r%s%s%s" % (
            props.get("BRANCH"),
            props.get("REVISION"),
            "M" if mods>0 else "",
            " (%s)" % commit if commit else "",
        )
        )


if __name__ == "__main__":
    main(sys.argv)
